import os
from itertools import product

import numpy as np
import pytest

from qililab import Domain, DragCorrection, Gaussian, IQPair, QProgram, Square
from qililab.qprogram.blocks import Average
from qililab.qprogram.calibration import Calibration
from qililab.qprogram.operations import (
    Acquire,
    AcquireWithCalibratedWeights,
    Measure,
    MeasureWithCalibratedWaveform,
    MeasureWithCalibratedWaveformWeights,
    MeasureWithCalibratedWeights,
    Play,
    PlayWithCalibratedWaveform,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetMarkers,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.utils.serialization import deserialize, deserialize_from, serialize, serialize_to
from tests.qprogram.test_structured_program import (
    TestStructuredProgram,
)


@pytest.fixture(name="sample_qprogram_string")
def get_sample_qprogram_string():
    """Sample qprogram and its corresponding string to tests the __str__ method"""
    r_amp = 0.5
    r_duration = 40
    d_duration = 40

    r_wf_I = Square(amplitude=r_amp, duration=r_duration)
    r_wf_Q = Square(amplitude=0.0, duration=r_duration)
    d_wf = IQPair.DRAG(amplitude=1.0, duration=d_duration, num_sigmas=4, drag_coefficient=0.1)

    weights_shape = Square(amplitude=1, duration=r_duration)

    qp = QProgram()
    amp = qp.variable(label="amplitude", domain=Domain.Voltage)
    freq = qp.variable(label="frequency", domain=Domain.Frequency)

    with qp.average(100):
        with qp.for_loop(variable=amp, start=0.2, stop=1, step=0.1):
            qp.set_gain(bus="dummy_bus_0", gain=amp)
            with qp.for_loop(variable=freq, start=0, stop=20, step=5):
                qp.set_frequency(bus="dummy_bus_1", frequency=freq)
                # DRAG PULSE
                qp.play(bus="dummy_bus_0", waveform=d_wf)
                qp.sync()
                # READOUT PULSE
                qp.measure(
                    bus="readout", waveform=IQPair(I=r_wf_I, Q=r_wf_Q), weights=IQPair(I=weights_shape, Q=weights_shape)
                )
                qp.wait(bus="readout", duration=200)

    qp_string = """Average:\n\tshots: 100\n\tForLoop:\n\t\tstart: 0.2\n\t\tstop: 1\n\t\tstep: 0.1\n\t\tSetGain:\n\t\t\tbus: dummy_bus_0\n\t\t\tgain: None\n\t\tForLoop:\n\t\t\tstart: 0\n\t\t\tstop: 20\n\t\t\tstep: 5\n\t\t\tSetFrequency:\n\t\t\t\tbus: dummy_bus_1\n\t\t\t\tfrequency: None\n\t\t\tPlay:\n\t\t\t\tbus: dummy_bus_0\n\t\t\t\twait_time: None\n\t\t\t\tWaveform I Gaussian:\n\t\t\t\t\t[0.         0.03369997 0.07235569 0.11612685 0.1650374  0.21894866\n\t\t\t\t\t 0.27753626 0.34027302 0.40641993 0.47502707 0.54494577 0.61485281\n\t\t\t\t\t 0.68328653 0.74869396 0.80948709 0.86410559 0.9110827  0.94911031\n\t\t\t\t\t 0.97709942 0.99423184 1.         0.99423184 0.97709942 0.94911031\n\t\t\t\t\t 0.9110827  0.86410559 0.80948709 0.74869396 0.68328653 0.61485281\n\t\t\t\t\t 0.54494577 0.47502707 0.40641993 0.34027302 0.27753626 0.21894866\n\t\t\t\t\t 0.1650374  0.11612685 0.07235569 0.03369997]\n\t\t\t\tWaveform Q DragCorrection):\n\t\t\t\t\t[ 0.          0.0006403   0.0013024   0.00197416  0.0026406   0.00328423\n\t\t\t\t\t  0.00388551  0.00442355  0.00487704  0.0052253   0.00544946  0.00553368\n\t\t\t\t\t  0.00546629  0.00524086  0.00485692  0.00432053  0.00364433  0.00284733\n\t\t\t\t\t  0.0019542   0.00099423 -0.         -0.00099423 -0.0019542  -0.00284733\n\t\t\t\t\t -0.00364433 -0.00432053 -0.00485692 -0.00524086 -0.00546629 -0.00553368\n\t\t\t\t\t -0.00544946 -0.0052253  -0.00487704 -0.00442355 -0.00388551 -0.00328423\n\t\t\t\t\t -0.0026406  -0.00197416 -0.0013024  -0.0006403 ]\n\t\t\tSync:\n\t\t\t\tbuses: None\n\t\t\tMeasure:\n\t\t\t\tbus: readout\n\t\t\t\tsave_adc: False\n\t\t\t\trotation: None\n\t\t\t\tdemodulation: True\n\t\t\t\tWaveform I Square:\n\t\t\t\t\t[0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5\n\t\t\t\t\t 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5 0.5\n\t\t\t\t\t 0.5 0.5 0.5 0.5]\n\t\t\t\tWaveform Q Square):\n\t\t\t\t\t[0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.\n\t\t\t\t\t 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0. 0.]\n\t\t\t\tWeights I Square:\n\t\t\t\t\t[1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1.\n\t\t\t\t\t 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1.]\n\t\t\t\tWeights Q Square:\n\t\t\t\t\t[1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1.\n\t\t\t\t\t 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1. 1.]\n\t\t\tWait:\n\t\t\t\tbus: readout\n\t\t\t\tduration: 200\n"""
    return (qp, qp_string)


class TestQProgram(TestStructuredProgram):
    """Unit tests checking the QProgram attributes and methods"""

    @pytest.fixture
    def instance(self):
        return QProgram()

    def test_with_bus_mapping_method(self):
        """Test with_bus_mapping method"""
        qp = QProgram()
        with qp.average(1000):
            qp.wait(bus="drive_bus", duration=100)
            qp.sync(buses=["drive_bus", "readout_bus"])
            qp.play(bus="drive_bus", waveform=Square(1.0, 100))

        new_qp = qp.with_bus_mapping(bus_mapping={"drive_bus": "drive_q0_bus", "readout_bus": "readout_q0_bus"})

        assert len(new_qp.buses) == 2
        assert "drive_bus" not in new_qp.buses
        assert "readout_bus" not in new_qp.buses
        assert "drive_q0_bus" in new_qp.buses
        assert "readout_q0_bus" in new_qp.buses

        assert new_qp.body.elements[0].elements[0].bus == "drive_q0_bus"
        assert new_qp.body.elements[0].elements[1].buses[0] == "drive_q0_bus"
        assert new_qp.body.elements[0].elements[1].buses[1] == "readout_q0_bus"
        assert new_qp.body.elements[0].elements[2].bus == "drive_q0_bus"

        self_mapping_qp = qp.with_bus_mapping(bus_mapping={"drive_bus": "drive_bus", "readout_bus": "readout_bus"})

        assert len(self_mapping_qp.buses) == 2
        assert "drive_bus" in self_mapping_qp.buses
        assert "readout_bus" in self_mapping_qp.buses

        assert self_mapping_qp.body.elements[0].elements[0].bus == "drive_bus"
        assert self_mapping_qp.body.elements[0].elements[1].buses[0] == "drive_bus"
        assert self_mapping_qp.body.elements[0].elements[1].buses[1] == "readout_bus"
        assert self_mapping_qp.body.elements[0].elements[2].bus == "drive_bus"

        non_existant_mapping_qp = qp.with_bus_mapping(
            bus_mapping={"non_existant": "drive_bus", "non_existant_readout": "readout_bus"}
        )

        assert len(non_existant_mapping_qp.buses) == 2
        assert "drive_bus" in non_existant_mapping_qp.buses
        assert "readout_bus" in non_existant_mapping_qp.buses

        assert non_existant_mapping_qp.body.elements[0].elements[0].bus == "drive_bus"
        assert non_existant_mapping_qp.body.elements[0].elements[1].buses[0] == "drive_bus"
        assert non_existant_mapping_qp.body.elements[0].elements[1].buses[1] == "readout_bus"
        assert non_existant_mapping_qp.body.elements[0].elements[2].bus == "drive_bus"

    def test_with_calibration_method(self):
        """Test with_bus_mapping method"""
        xgate = IQPair.DRAG(amplitude=1.0, duration=40, num_sigmas=4.5, drag_coefficient=-4.5)
        readout = IQPair(I=Square(1.0, 200), Q=Square(1.0, 200))
        weights = IQPair(I=Square(1.0, 2000), Q=Square(1.0, 2000))

        calibration = Calibration()
        calibration.add_waveform(bus="drive_q0_bus", name="xgate", waveform=xgate)
        calibration.add_waveform(bus="readout_q0_bus", name="readout", waveform=readout)
        calibration.add_weights(bus="readout_q0_bus", name="weights", weights=weights)

        qp = QProgram()
        with qp.average(1000):
            qp.play(bus="drive_q0_bus", waveform="xgate")
            qp.qblox.play(bus="drive_q0_bus", waveform="xgate", wait_time=100)
            qp.qblox.acquire(bus="readout_q0_bus", weights="weights")
            qp.measure(bus="readout_q0_bus", waveform="readout", weights=weights)
            qp.measure(bus="readout_q0_bus", waveform=readout, weights="weights")
            qp.measure(bus="readout_q0_bus", waveform="readout", weights="weights")
            qp.quantum_machines.measure(bus="readout_q0_bus", waveform="readout", weights=weights, rotation=np.pi)
            qp.quantum_machines.measure(bus="readout_q0_bus", waveform=readout, weights="weights", rotation=np.pi)
            qp.quantum_machines.measure(bus="readout_q0_bus", waveform="readout", weights="weights", rotation=np.pi)

        # Check that qp has named operations
        assert qp.has_calibrated_waveforms_or_weights() is True
        assert isinstance(qp.body.elements[0].elements[0], PlayWithCalibratedWaveform)
        assert qp.body.elements[0].elements[0].waveform == "xgate"
        assert isinstance(qp.body.elements[0].elements[1], PlayWithCalibratedWaveform)
        assert qp.body.elements[0].elements[1].waveform == "xgate"
        assert isinstance(qp.body.elements[0].elements[2], AcquireWithCalibratedWeights)
        assert qp.body.elements[0].elements[2].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[3], MeasureWithCalibratedWaveform)
        assert qp.body.elements[0].elements[3].waveform == "readout"
        assert isinstance(qp.body.elements[0].elements[4], MeasureWithCalibratedWeights)
        assert qp.body.elements[0].elements[4].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[5], MeasureWithCalibratedWaveformWeights)
        assert qp.body.elements[0].elements[5].waveform == "readout"
        assert qp.body.elements[0].elements[5].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[6], MeasureWithCalibratedWaveform)
        assert qp.body.elements[0].elements[6].waveform == "readout"
        assert isinstance(qp.body.elements[0].elements[7], MeasureWithCalibratedWeights)
        assert qp.body.elements[0].elements[7].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[8], MeasureWithCalibratedWaveformWeights)
        assert qp.body.elements[0].elements[8].waveform == "readout"
        assert qp.body.elements[0].elements[8].weights == "weights"

        new_qp = qp.with_calibration(calibration=calibration)

        # Check that qp remain unchanged
        assert qp.has_calibrated_waveforms_or_weights() is True
        assert isinstance(qp.body.elements[0].elements[0], PlayWithCalibratedWaveform)
        assert qp.body.elements[0].elements[0].waveform == "xgate"
        assert isinstance(qp.body.elements[0].elements[1], PlayWithCalibratedWaveform)
        assert qp.body.elements[0].elements[1].waveform == "xgate"
        assert isinstance(qp.body.elements[0].elements[2], AcquireWithCalibratedWeights)
        assert qp.body.elements[0].elements[2].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[3], MeasureWithCalibratedWaveform)
        assert qp.body.elements[0].elements[3].waveform == "readout"
        assert isinstance(qp.body.elements[0].elements[4], MeasureWithCalibratedWeights)
        assert qp.body.elements[0].elements[4].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[5], MeasureWithCalibratedWaveformWeights)
        assert qp.body.elements[0].elements[5].waveform == "readout"
        assert qp.body.elements[0].elements[5].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[6], MeasureWithCalibratedWaveform)
        assert qp.body.elements[0].elements[6].waveform == "readout"
        assert isinstance(qp.body.elements[0].elements[7], MeasureWithCalibratedWeights)
        assert qp.body.elements[0].elements[7].weights == "weights"
        assert isinstance(qp.body.elements[0].elements[8], MeasureWithCalibratedWaveformWeights)
        assert qp.body.elements[0].elements[8].waveform == "readout"
        assert qp.body.elements[0].elements[8].weights == "weights"

        # Check that new_qp has no named operations
        assert new_qp.has_calibrated_waveforms_or_weights() is False
        assert isinstance(new_qp.body.elements[0].elements[0], Play)
        assert isinstance(new_qp.body.elements[0].elements[1], Play)
        assert isinstance(new_qp.body.elements[0].elements[2], Acquire)
        assert isinstance(new_qp.body.elements[0].elements[3], Measure)
        assert isinstance(new_qp.body.elements[0].elements[4], Measure)
        assert isinstance(new_qp.body.elements[0].elements[5], Measure)
        assert isinstance(new_qp.body.elements[0].elements[6], Measure)
        assert isinstance(new_qp.body.elements[0].elements[7], Measure)
        assert isinstance(new_qp.body.elements[0].elements[8], Measure)

    def test_average_method(self):
        """Test acquire_loop method"""
        qp = QProgram()
        with qp.average(shots=1000) as loop:
            # __enter__
            assert isinstance(loop, Average)
            assert loop.shots == 1000
            assert qp._active_block is loop
        # __exit__
        assert len(qp._body.elements) == 1
        assert qp._body.elements[0] is loop
        assert qp._active_block is qp._body

    def test_play_method(self):
        """Test play method"""
        one_wf = Square(amplitude=1.0, duration=40)
        zero_wf = Square(amplitude=0.0, duration=40)
        qp = QProgram()
        qp.play(bus="flux", waveform=one_wf)
        qp.play(bus="drive", waveform=IQPair(I=one_wf, Q=zero_wf))

        assert len(qp._active_block.elements) == 2
        assert len(qp._body.elements) == 2
        assert isinstance(qp._body.elements[0], Play)
        assert qp._body.elements[0].bus == "flux"
        assert np.equal(qp._body.elements[0].waveform, one_wf)
        assert isinstance(qp._body.elements[1], Play)
        assert qp._body.elements[1].bus == "drive"
        assert np.equal(qp._body.elements[1].waveform.I, one_wf)
        assert np.equal(qp._body.elements[1].waveform.Q, zero_wf)

    def test_wait_method(self):
        """Test wait method"""
        qp = QProgram()
        qp.wait(bus="drive", duration=100)

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], Wait)
        assert qp._body.elements[0].bus == "drive"
        assert qp._body.elements[0].duration == 100

    def test_measure_method(self):
        """Test measure method"""
        one_wf = Square(amplitude=1.0, duration=40)
        zero_wf = Square(amplitude=0.0, duration=40)
        qp = QProgram()
        qp.measure(bus="readout", waveform=IQPair(one_wf, zero_wf), weights=IQPair(one_wf, zero_wf))

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], Measure)
        assert qp._body.elements[0].bus == "readout"
        assert np.equal(qp._body.elements[0].waveform.I, one_wf)
        assert np.equal(qp._body.elements[0].waveform.Q, zero_wf)
        assert np.equal(qp._body.elements[0].weights.I, one_wf)
        assert np.equal(qp._body.elements[0].weights.Q, zero_wf)

    def test_acquire_method(self):
        """Test acquire method"""
        one_wf = Square(amplitude=1.0, duration=40)
        zero_wf = Square(amplitude=0.0, duration=40)
        qp = QProgram()
        qp.qblox.acquire(bus="readout", weights=IQPair(I=one_wf, Q=zero_wf))

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], Acquire)
        assert qp._body.elements[0].bus == "readout"
        assert np.equal(qp._body.elements[0].weights.I, one_wf)
        assert np.equal(qp._body.elements[0].weights.Q, zero_wf)

    def test_sync_method(self):
        """Test sync method"""
        qp = QProgram()
        qp.sync(buses=["drive", "flux"])

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], Sync)
        assert qp._body.elements[0].buses == ["drive", "flux"]

    def test_reset_phase(self):
        """Test reset_nco_phase method"""
        qp = QProgram()
        qp.reset_phase(bus="drive")

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], ResetPhase)
        assert qp._body.elements[0].bus == "drive"

    def test_set_phase(self):
        """Test set_nco_phase method"""
        qp = QProgram()
        qp.set_phase(bus="drive", phase=2 * np.pi)

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], SetPhase)
        assert qp._body.elements[0].bus == "drive"
        assert qp._body.elements[0].phase == 2 * np.pi

    def test_set_frequency(self):
        """Test set_nco_frequency method"""
        qp = QProgram()
        qp.set_frequency(bus="drive", frequency=1e6)

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], SetFrequency)
        assert qp._body.elements[0].bus == "drive"
        assert qp._body.elements[0].frequency == 1e6

    def test_set_gain(self):
        """Test set_awg_gain method"""
        qp = QProgram()
        qp.set_gain(bus="drive", gain=1.0)

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], SetGain)
        assert qp._body.elements[0].bus == "drive"
        assert qp._body.elements[0].gain == 1.0

    def test_set_offset(self):
        """Test set_awg_offset method"""
        qp = QProgram()
        qp.set_offset(bus="drive", offset_path0=1.0, offset_path1=0.0)

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], SetOffset)
        assert qp._body.elements[0].bus == "drive"
        assert qp._body.elements[0].offset_path0 == 1.0
        assert qp._body.elements[0].offset_path1 == 0.0

    def test_set_markers(self):
        qp = QProgram()
        qp.qblox.set_markers(bus="drive", mask="0111")

        assert len(qp._active_block.elements) == 1
        assert len(qp._body.elements) == 1
        assert isinstance(qp._body.elements[0], SetMarkers)
        assert qp._body.elements[0].bus == "drive"
        assert qp._body.elements[0].mask == "0111"

        with pytest.raises(AttributeError):
            qp.qblox.set_markers(bus="drive", mask="1234")

        with pytest.raises(AttributeError):
            qp.qblox.set_markers(bus="drive", mask="0111011")

    def test_operation_with_variable_of_wrong_domain_raises_error(self):
        """Test that any operation when used with a variable of wrong domain raises an error."""
        qp = QProgram()
        frequency = qp.variable(label="frequency", domain=Domain.Frequency)
        phase = qp.variable(label="phase", domain=Domain.Phase)
        voltage = qp.variable(label="gain", domain=Domain.Voltage)
        time = qp.variable(label="time", domain=Domain.Time)
        scalar = qp.variable(label="float_scalar", domain=Domain.Scalar, type=float)

        all_types = {frequency, phase, voltage, time, scalar}

        for var in all_types - {frequency}:
            with pytest.raises(ValueError):
                qp.set_frequency(bus="drive", frequency=var)

        for var in all_types - {phase}:
            with pytest.raises(ValueError):
                qp.set_phase(bus="drive", phase=var)

        for var in all_types - {voltage}:
            with pytest.raises(ValueError):
                qp.set_offset(bus="drive", offset_path0=var, offset_path1=var)

        for var in all_types - {voltage}:
            with pytest.raises(ValueError):
                qp.set_gain(bus="drive", gain=var)

        for var in all_types - {time}:
            with pytest.raises(ValueError):
                qp.wait(bus="drive", duration=var)

        for amplitude_var, duration_var in set(product(all_types, repeat=2)) - {(voltage, time)}:
            with pytest.raises(ValueError):
                _ = Square(amplitude=amplitude_var, duration=duration_var)

        for amplitude_var, duration_var, num_sigmas_var in set(product(all_types, repeat=3)) - {
            (voltage, time, scalar)
        }:
            with pytest.raises(ValueError):
                _ = Gaussian(amplitude=amplitude_var, duration=duration_var, num_sigmas=num_sigmas_var)

        for var in all_types - {scalar}:
            gaussian = Gaussian(amplitude=1.0, duration=40, num_sigmas=2.5)
            with pytest.raises(ValueError):
                _ = DragCorrection(drag_coefficient=var, waveform=gaussian)

        for amplitude_var, duration_var, num_sigmas_var, drag_coefficient_var in set(product(all_types, repeat=4)) - {
            (voltage, time, scalar, scalar)
        }:
            with pytest.raises(ValueError):
                _ = IQPair.DRAG(
                    amplitude=amplitude_var,
                    duration=duration_var,
                    num_sigmas=num_sigmas_var,
                    drag_coefficient=drag_coefficient_var,
                )

    def test_serialization_deserialization(self):
        """Test serialization and deserialization works."""
        file = "test_serialization_deserialization_qprogram.yml"
        qp = QProgram()
        gain = qp.variable(label="gain", domain=Domain.Voltage)
        with qp.for_loop(variable=gain, start=0.0, stop=1.0, step=0.1):
            qp.set_gain(bus="drive_bus", gain=gain)
            qp.play(bus="drive_bus", waveform=IQPair(I=Square(1.0, 200), Q=Square(1.0, 200)))

        serialized = serialize(qp)
        deserialized_qprogram = deserialize(serialized, QProgram)

        assert isinstance(deserialized_qprogram, QProgram)

        serialize_to(qp, file=file)
        deserialized_qprogram = deserialize_from(file, QProgram)

        assert isinstance(deserialized_qprogram, QProgram)

        os.remove(file)



import numpy as np
import pytest

from qililab.waveforms import Gaussian, Square
from qililab.waveforms.drag_correction import DragCorrection


@pytest.fixture(name="gaussian")
def fixture_gaussian():
    return Gaussian(amplitude=1, duration=10, num_sigmas=2.5)


class TestDragCorrection:
    def test_init(self, gaussian):
        # test init method
        drag_correction = DragCorrection(0.8, gaussian)
        assert drag_correction.drag_coefficient == 0.8
        assert isinstance(drag_correction.waveform, Gaussian)

    def test_envelope_gaussian(self, gaussian):
        # test envelope method

        drag_correction = DragCorrection(0.8, gaussian)
        sigma = gaussian.duration / gaussian.num_sigmas
        mu = gaussian.duration / 2
        x = np.arange(gaussian.duration)
        envelope = (-0.8 * (x - mu) / sigma**2) * gaussian.envelope()
        assert np.allclose(drag_correction.envelope(), envelope)

    def test_envelope_not_implemented_error(self):
        square = Square(amplitude=0.5, duration=100)
        drag_correction = DragCorrection(0.8, square)
        error_string = "Cannot apply drag correction on a Square waveform."
        with pytest.raises(NotImplementedError, match=error_string):
            drag_correction.envelope()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Quantum Machines Manager Instrument Controller """
from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentControllerSettings
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.quantum_machines import QuantumMachinesCluster
from qililab.typings import QMMDriver
from qililab.typings.enums import InstrumentControllerName


@InstrumentControllerFactory.register
class QuantumMachinesClusterController(SingleInstrumentController):
    """Quantum Machines Manager class.

    This class implements the instrument controller for the Quantum Machines Manager instrument wrapper in Qililab.

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        device (QMMDriver): Instance of the Quantum Machines Manager Driver class.
        settings (QMMControllerSettings): Settings of the instrument.
    """

    name = InstrumentControllerName.QUANTUM_MACHINES_CLUSTER
    device: QMMDriver
    modules: Sequence[QuantumMachinesCluster]

    @dataclass
    class QuantumMachinesClusterControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific Quantum Machines Manager Controller."""

    settings: QuantumMachinesClusterControllerSettings

    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""
        self.device = QMMDriver()

    def _check_supported_modules(self):
        """Checks if all instrument modules loaded are supported modules for the controller."""

    def _set_device_to_all_modules(self):
        """Sets the initialized device to all modules."""
        for module in self.modules:
            module.device = self.device



# pylint: disable=protected-access
import os
from datetime import datetime
from pathlib import Path
from types import MethodType
from unittest.mock import MagicMock, create_autospec, patch

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest

from qililab.result.experiment_results import DimensionInfo, ExperimentResults
from qililab.result.experiment_results_writer import ExperimentMetadata, ExperimentResultsWriter

mpl.use("Agg")  # Use non-interactive backend for testing


# Dummy path for testing
EXPERIMENT_RESULTS_PATH = "dummy.hdf5"


@pytest.fixture(name="metadata")
def sample_metadata():
    """Create a mock metadata dictionary for testing"""
    return ExperimentMetadata(
        experiment="experiment",
        platform="platform",
        executed_at=datetime(2024, 1, 1, 0, 0, 0, 0),
        execution_time=1.23,
        qprograms={
            "QProgram_0": {
                "variables": [{"label": "x", "values": np.array([1, 2, 3])}],
                "dims": [["x"]],
                "measurements": {
                    # This simulates a single loop
                    "Measurement_0": {
                        "variables": [{"label": "y", "values": np.array([10, 20, 30, 40])}],
                        "dims": [["y"]],
                        "shape": (3, 4, 2),
                        "shots": 100,
                    },
                    # This simulates an inner loop
                    "Measurement_1": {
                        "variables": [
                            {"label": "y", "values": np.array([10, 20, 30, 40])},
                            {"label": "z", "values": np.array([110, 120, 130, 140])},
                        ],
                        "dims": [["y"], ["z"]],
                        "shape": (3, 4, 4, 2),
                        "shots": 100,
                    },
                    # This simulates a parallel loop
                    "Measurement_2": {
                        "variables": [
                            {"label": "y", "values": np.array([10, 20, 30, 40])},
                            {"label": "z", "values": np.array([110, 120, 130, 140])},
                        ],
                        "dims": [["y", "z"]],
                        "shape": (3, 4, 2),
                        "shots": 100,
                    },
                },
            }
        },
    )


@pytest.fixture(name="experiment_results")
def mock_experiment_results(metadata):
    """Create a mock HDF5 file structure for testing"""
    with ExperimentResultsWriter(path=EXPERIMENT_RESULTS_PATH, metadata=metadata):
        ...
    yield EXPERIMENT_RESULTS_PATH
    Path(EXPERIMENT_RESULTS_PATH).unlink()


class TestExperimentResults:
    """Test ExperimentResults class"""

    def test_enter_exit(self, experiment_results):
        """Test __enter__ and __exit__"""
        # Test that opening and closing the file works as expected
        with ExperimentResults(experiment_results) as exp_results:
            assert exp_results._file is not None
            assert exp_results._file.id.valid
        # After exiting the context manager, the file should be closed
        assert not exp_results._file.id.valid

    def test_get_item(self, experiment_results):
        """Test __get_item__"""
        with ExperimentResults(experiment_results) as exp_results:
            value = exp_results["QProgram_0", "Measurement_0", 0, 0, 0]
            assert value == 0.0

    def test_len(self, experiment_results):
        """Test __len__"""
        with ExperimentResults(experiment_results) as exp_results:
            assert len(exp_results) == 3

    def test_iter(self, experiment_results):
        """Test __iter__"""
        with ExperimentResults(experiment_results) as exp_results:
            results = list(iter(exp_results))
            assert len(results) == 3
            assert results[0][0] == ("QProgram_0", "Measurement_0")

    def test_properties(self, experiment_results):
        """Test properties"""
        with ExperimentResults(experiment_results) as exp_results:
            assert exp_results.experiment == "experiment"
            assert exp_results.platform == "platform"
            assert exp_results.executed_at == datetime(2024, 1, 1, 0, 0, 0, 0)

    def test_get_method(self, experiment_results):
        """Test get method"""
        with ExperimentResults(experiment_results) as exp_results:
            data, dims = exp_results.get()
            assert data.shape == (3, 4, 2)
            assert len(dims) == 3
            assert dims[0].labels[0] == "x"
            assert dims[1].labels[0] == "y"
            assert dims[2].labels[0] == "I/Q"

            data, dims = exp_results.get(qprogram=0, measurement=0)
            assert data.shape == (3, 4, 2)
            assert len(dims) == 3
            assert dims[0].labels[0] == "x"
            assert dims[1].labels[0] == "y"
            assert dims[2].labels[0] == "I/Q"

            data, dims = exp_results.get(qprogram="QProgram_0", measurement="Measurement_0")
            assert data.shape == (3, 4, 2)
            assert len(dims) == 3
            assert dims[0].labels[0] == "x"
            assert dims[1].labels[0] == "y"
            assert dims[2].labels[0] == "I/Q"

    def test_plot_S21_one_dimension(self):
        """Test plot_S21 with 1D data and verify the plot correctness."""
        experiment_results = create_autospec(ExperimentResults, instance=True, path="test.h5")

        # Manually set the plot_S21 method to the real one
        experiment_results.plot_S21 = MethodType(ExperimentResults.plot_S21, experiment_results)

        # Mock the get method to return 1D data
        data = np.random.rand(100, 2)  # 100 data points, real and imaginary parts
        dims = [
            DimensionInfo(labels=["Frequency (Hz)"], values=[np.linspace(1e6, 1e7, 100)]),
            DimensionInfo(labels=["I/Q"], values=[]),
        ]

        experiment_results.get.return_value = (data, dims)

        # Call the plot_S21 method
        experiment_results.plot_S21()

        # Ensure get was called correctly
        experiment_results.get.assert_called_with(qprogram=0, measurement=0)

        # Retrieve the current figure and axes
        fig = plt.gcf()
        ax1 = fig.axes[0]

        # Check the title, labels, and line data
        assert ax1.get_title() == experiment_results.path
        assert ax1.get_xlabel() == "Frequency (Hz)"
        assert ax1.get_ylabel() == r"$|S_{21}|$"

        lines = ax1.get_lines()
        assert len(lines) == 1  # Should have one line plotted

        # Verify the data plotted
        x_plotted = lines[0].get_xdata()
        y_plotted = lines[0].get_ydata()

        # Recompute s21 to compare
        s21 = data[:, 0] + 1j * data[:, 1]
        s21_db = 20 * np.log10(np.abs(s21))
        x_expected = dims[0].values[0]

        np.testing.assert_array_almost_equal(x_plotted, x_expected)
        np.testing.assert_array_almost_equal(y_plotted, s21_db)

        # Close the plot
        plt.close(fig)

        # Verify that the plot was saved to a file.
        plot_path = os.path.join(os.path.dirname(experiment_results.path), ExperimentResults.S21_PLOT_NAME)
        assert os.path.exists(plot_path)

        # Remove the file
        os.remove(plot_path)

        # Verify that if `save_plot == False` then the plot is not saved.
        experiment_results.plot_S21(save_plot=False)
        assert not os.path.exists(plot_path)

        # Close the plot
        plt.close(plt.gcf())

    def test_plot_S21_one_dimension_secondary_axis(self):
        """Test plot_S21 with 1D data and secondary x-axis."""
        experiment_results = create_autospec(ExperimentResults, instance=True, path="test.h5")

        # Manually set the plot_S21 method to the real one
        experiment_results.plot_S21 = MethodType(ExperimentResults.plot_S21, experiment_results)

        # Mock the get method to return 1D data with secondary axis data
        data = np.random.rand(100, 2)  # 100 data points, real and imaginary parts
        dims = [
            DimensionInfo(
                labels=["Frequency (Hz)", "Time (s)"], values=[np.linspace(1e6, 1e7, 100), np.linspace(0, 1, 100)]
            ),
            DimensionInfo(labels=["I/Q"], values=[]),
        ]

        experiment_results.get.return_value = (data, dims)

        # Call the plot_S21 method
        experiment_results.plot_S21()

        # Retrieve the current figure and axes
        fig = plt.gcf()
        ax1 = fig.axes[0]
        ax2 = fig.axes[1]

        # Verify that the secondary axis is created
        assert ax2 is not None, "Secondary x-axis was not created."

        # Check labels
        assert ax1.get_xlabel() == "Frequency (Hz)"
        assert ax2.get_xlabel() == "Time (s)"

        # Check limits
        np.testing.assert_almost_equal(ax2.get_xlim(), [0, 1])

        # Check ticks
        expected_ticks = np.linspace(0, 1, num=6)
        np.testing.assert_array_almost_equal(ax2.get_xticks(), expected_ticks)

        # Close the plot
        plt.close(fig)

        # Verify that the plot was saved to a file.
        plot_path = os.path.join(os.path.dirname(experiment_results.path), ExperimentResults.S21_PLOT_NAME)
        assert os.path.exists(plot_path)

        # Remove the file
        os.remove(plot_path)

        # Verify that if `save_plot == False` then the plot is not saved.
        experiment_results.plot_S21(save_plot=False)
        assert not os.path.exists(plot_path)

        # Close the plot
        plt.close(plt.gcf())

    def test_plot_S21_two_dimensions(self):
        """Test plot_S21 with 2D data and verify the plot correctness."""
        experiment_results = create_autospec(ExperimentResults, instance=True, path="test.h5")

        # Manually set the plot_S21 method to the real one
        experiment_results.plot_S21 = MethodType(ExperimentResults.plot_S21, experiment_results)

        # Mock the get method to return 2D data
        data = np.random.rand(50, 50, 2)  # 50x50 data points, real and imaginary parts
        dims = [
            DimensionInfo(labels=["Frequency (Hz)"], values=[np.linspace(1e6, 1e7, 50)]),
            DimensionInfo(labels=["Voltage (V)"], values=[np.linspace(0, 1, 50)]),
            DimensionInfo(labels=["I/Q"], values=[]),
        ]

        experiment_results.get.return_value = (data, dims)

        # Call the plot_S21 method
        experiment_results.plot_S21()

        # Ensure get was called correctly
        experiment_results.get.assert_called_with(qprogram=0, measurement=0)

        # Retrieve the current figure and axes
        fig = plt.gcf()
        ax1 = fig.axes[0]
        colorbar = fig.axes[1]

        # Check the title and labels
        assert ax1.get_title() == experiment_results.path
        assert ax1.get_xlabel() == "Frequency (Hz)"
        assert ax1.get_ylabel() == "Voltage (V)"

        # Retrieve the QuadMesh object from pcolormesh
        quadmesh = ax1.collections[0]
        assert isinstance(quadmesh, mpl.collections.QuadMesh)

        # Verify the data plotted
        s21 = data[:, :, 0] + 1j * data[:, :, 1]
        s21_db = 20 * np.log10(np.abs(s21))

        # Get the mesh data
        mesh_array = quadmesh.get_array().reshape(s21_db.T.shape)

        # Compare the mesh data to s21_db
        np.testing.assert_array_almost_equal(mesh_array, s21_db.T)

        # Verify that the colorbar axes exist and have correct properties
        assert colorbar is not None, "Colorbar axes was not created."

        # Colorbar axes usually have no labels
        assert not colorbar.get_xlabel(), "Colorbar axes should not have an x-label."
        assert not colorbar.get_ylabel(), "Colorbar axes should not have a y-label."

        # Close the plot
        plt.close(fig)

        # Verify that the plot was saved to a file.
        plot_path = os.path.join(os.path.dirname(experiment_results.path), ExperimentResults.S21_PLOT_NAME)
        assert os.path.exists(plot_path)

        # Remove the file
        os.remove(plot_path)

        # Verify that if `save_plot == False` then the plot is not saved.
        experiment_results.plot_S21(save_plot=False)
        assert not os.path.exists(plot_path)

        # Close the plot
        plt.close(plt.gcf())

    def test_plot_S21_two_dimensions_secondary_axes(self):
        """Test plot_S21 with 2D data and secondary axes."""
        experiment_results = create_autospec(ExperimentResults, instance=True, path="test.h5")

        # Manually set the plot_S21 method to the real one
        experiment_results.plot_S21 = MethodType(ExperimentResults.plot_S21, experiment_results)

        # Mock the get method to return 2D data with secondary axis data
        data = np.random.rand(50, 50, 2)  # 50x50 data points, real and imaginary parts
        dims = [
            DimensionInfo(
                labels=["Frequency (Hz)", "Wavelength (m)"],
                values=[np.linspace(1e6, 1e7, 50), np.linspace(300e-9, 800e-9, 50)],
            ),
            DimensionInfo(
                labels=["Voltage (V)", "Current (A)"], values=[np.linspace(0, 1, 50), np.linspace(0, 10, 50)]
            ),
            DimensionInfo(labels=["I/Q"], values=[]),
        ]

        experiment_results.get.return_value = (data, dims)

        # Call the plot_S21 method
        experiment_results.plot_S21()

        # Retrieve the current figure and axes
        fig = plt.gcf()
        ax1 = fig.axes[0]
        colorbar = fig.axes[1]
        ax2 = fig.axes[2]
        ax3 = fig.axes[3]

        # Verify that the secondary axes are created
        assert ax2 is not None, "Secondary x-axis was not created."
        assert ax3 is not None, "Secondary y-axis was not created."

        # Check labels
        assert ax1.get_xlabel() == "Frequency (Hz)"
        assert ax1.get_ylabel() == "Voltage (V)"
        assert ax2.get_xlabel() == "Wavelength (m)"
        assert ax3.get_ylabel() == "Current (A)"

        # Check limits
        np.testing.assert_almost_equal(ax2.get_xlim(), [300e-9, 800e-9])
        np.testing.assert_almost_equal(ax3.get_ylim(), [0, 10])

        # Check ticks for ax2 (secondary x-axis)
        expected_xticks = np.linspace(300e-9, 800e-9, num=6)
        np.testing.assert_array_almost_equal(ax2.get_xticks(), expected_xticks)

        # Check ticks for ax3 (secondary y-axis)
        expected_yticks = np.linspace(0, 10, num=6)
        np.testing.assert_array_almost_equal(ax3.get_yticks(), expected_yticks)

        # Verify that the colorbar axes exist and have correct properties
        assert colorbar is not None, "Colorbar axes was not created."

        # Colorbar axes usually have no labels
        assert not colorbar.get_xlabel(), "Colorbar axes should not have an x-label."
        assert not colorbar.get_ylabel(), "Colorbar axes should not have a y-label."

        # Close the plot
        plt.close(fig)

        # Verify that the plot was saved to a file.
        plot_path = os.path.join(os.path.dirname(experiment_results.path), ExperimentResults.S21_PLOT_NAME)
        assert os.path.exists(plot_path)

        # Remove the file
        os.remove(plot_path)

        # Verify that if `save_plot == False` then the plot is not saved.
        experiment_results.plot_S21(save_plot=False)
        assert not os.path.exists(plot_path)

        # Close the plot
        plt.close(plt.gcf())

    def test_plot_S21_three_dimensions(self):
        """Test plot_S21 with 3D data, should raise NotImplementedError."""
        experiment_results = create_autospec(ExperimentResults, instance=True, path="test.h5")

        # Manually set the plot_S21 method to the real one
        experiment_results.plot_S21 = MethodType(ExperimentResults.plot_S21, experiment_results)

        # Mock the get method to return 3D data
        data = np.random.rand(10, 10, 10, 2)
        dims = [
            {"labels": ["Frequency (Hz)"], "values": [np.linspace(1e6, 1e7, 10)]},
            {"labels": ["Voltage (V)"], "values": [np.linspace(0, 1, 10)]},
            {"labels": ["Temperature (K)"], "values": [np.linspace(0, 300, 10)]},
            {"labels": ["I/Q"], "values": []},
        ]

        experiment_results.get.return_value = (data, dims)

        # Check that NotImplementedError is raised
        with pytest.raises(NotImplementedError):
            experiment_results.plot_S21()

        # Ensure get was called correctly
        experiment_results.get.assert_called_with(qprogram=0, measurement=0)

    def test_plot_S21_with_qprogram_and_measurement_strings(self):
        """Test plot_S21 with string qprogram and measurement and verify the plot correctness."""
        experiment_results = create_autospec(ExperimentResults, instance=True, path="test.h5")

        # Manually set the plot_S21 method to the real one
        experiment_results.plot_S21 = MethodType(ExperimentResults.plot_S21, experiment_results)

        # Mock the get method to return 1D data
        data = np.random.rand(100, 2)
        dims = [
            DimensionInfo(labels=["Frequency (Hz)"], values=[np.linspace(1e6, 1e7, 100)]),
            DimensionInfo(labels=["I/Q"], values=[]),
        ]

        experiment_results.get = MagicMock(return_value=(data, dims))

        # Call the plot_S21 method with string parameters
        qprogram = "program1"
        measurement = "measurement1"
        experiment_results.plot_S21(qprogram=qprogram, measurement=measurement)

        # Ensure get was called correctly
        experiment_results.get.assert_called_with(qprogram=qprogram, measurement=measurement)

        # Retrieve the current figure and axes
        fig = plt.gcf()
        ax1 = fig.axes[0]

        # Check the title, labels, and line data
        assert ax1.get_title() == experiment_results.path
        assert ax1.get_xlabel() == "Frequency (Hz)"
        assert ax1.get_ylabel() == r"$|S_{21}|$"

        lines = ax1.get_lines()
        assert len(lines) == 1  # Should have one line plotted

        # Verify the data plotted
        x_plotted = lines[0].get_xdata()
        y_plotted = lines[0].get_ydata()

        # Recompute s21 to compare
        s21 = data[:, 0] + 1j * data[:, 1]
        s21_db = 20 * np.log10(np.abs(s21))
        x_expected = dims[0].values[0]

        np.testing.assert_array_almost_equal(x_plotted, x_expected)
        np.testing.assert_array_almost_equal(y_plotted, s21_db)

        # Close the plot
        plt.close(fig)

        # Verify that the plot was saved to a file.
        plot_path = os.path.join(os.path.dirname(experiment_results.path), ExperimentResults.S21_PLOT_NAME)
        assert os.path.exists(plot_path)

        # Remove the file
        os.remove(plot_path)

        # Verify that if `save_plot == False` then the plot is not saved.
        experiment_results.plot_S21(save_plot=False)
        assert not os.path.exists(plot_path)

        # Close the plot
        plt.close(plt.gcf())


class TestExperimentResultsWriter:
    """Test ExperimentResultsWriter class"""

    @patch("h5py.File")
    def test_create_results_file(self, mock_h5file, metadata):
        """Test file creation"""
        # Test that the results file is created with the correct structure
        with ExperimentResultsWriter(path="mock_path", metadata=metadata):
            pass  # Just initializing should create the file structure

        assert mock_h5file.called
        mock_h5file.assert_called_with("mock_path", mode="w")

    def test_setters(self, experiment_results):
        """Test setters"""
        with ExperimentResultsWriter(path=experiment_results, metadata={}) as exp_writer:
            # test experiment property
            exp_writer.experiment = "new_experiment"
            assert exp_writer.experiment == "new_experiment"

            # write again to assert that HDF5 old partition is deleted correctly
            exp_writer.experiment = "newer_experiment"
            assert exp_writer.experiment == "newer_experiment"

            # test platform property
            exp_writer.platform = "new_platform"
            assert exp_writer.platform == "new_platform"

            # write again to assert that HDF5 old partition is deleted correctly
            exp_writer.platform = "newer_platform"
            assert exp_writer.platform == "newer_platform"

            # test crosstalk property
            exp_writer.crosstalk = "new_crosstalk"
            assert exp_writer.crosstalk == "new_crosstalk"

            # write again to assert that HDF5 old partition is deleted correctly
            exp_writer.crosstalk = "newer_crosstalk"
            assert exp_writer.crosstalk == "newer_crosstalk"

            # test executed_at property
            exp_writer.executed_at = datetime(2025, 1, 1, 0, 0, 0)
            assert exp_writer.executed_at == datetime(2025, 1, 1, 0, 0, 0)

            # write again to assert that HDF5 old partition is deleted correctly
            exp_writer.executed_at = datetime(2026, 1, 1, 0, 0, 0)
            assert exp_writer.executed_at == datetime(2026, 1, 1, 0, 0, 0)

            # test execution_time property
            exp_writer.execution_time = 1.23
            assert exp_writer.execution_time == 1.23

            # write again to assert that HDF5 old partition is deleted correctly
            exp_writer.execution_time = 4.56
            assert exp_writer.execution_time == 4.56



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantum Machines Manager class."""

import os
from dataclasses import dataclass
from typing import Any, cast

import numpy as np
from qm import DictQuaConfig, QmJob, QuantumMachine, QuantumMachinesManager, SimulationConfig
from qm.api.v2.job_api import JobApi
from qm.jobs.running_qm_job import RunningQmJob
from qm.octave import QmOctaveConfig
from qm.program import Program

from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue, QMMDriver
from qililab.utils import hash_qua_program, merge_dictionaries


@InstrumentFactory.register
class QuantumMachinesCluster(Instrument):
    """Class defining the Qililab Quantum Machines Cluster instrument in Qililab.

    This class allows Qililab control and communication with an instance of the
    Quantum Machines Manager, which is the central class to interact with Quantum Machines instruments.
    The manager is responsible, through the use of QUA sequences, of setting the Quantum Machines instruments
    in the right manner for Quantum Control.

    Args:
        name (InstrumentName): Name of the Instrument.
        device (QMMDriver): Instance of the Quantum Machines Manager Driver class.
        settings (QMMSettings): Settings of the instrument.
    """

    name = InstrumentName.QUANTUM_MACHINES_CLUSTER

    @dataclass
    class QuantumMachinesClusterSettings(Instrument.InstrumentSettings):
        """Settings for Quantum Machines Cluster instrument.

        Args:
            address (str): IP address to connect to the Quantum Machines instruments.
            cluster (str): Name of the cluster to connect to.
            run_octave_calibration (bool): Flag to indicate if octave calibration should be run on connecting to Quantum Machines.
            octaves (list[dict[str, Any]]): List of octaves the quantum machines stack has.
            controllers (list[dict[str, Any]]): List of controllers (instruments) the quantum machines stack has.
            elements (list[dict[str, Any]]): List of elements (buses) the quantum machines stack has.
        """

        address: str
        cluster: str
        run_octave_calibration: bool
        octaves: list[dict[str, Any]]
        controllers: list[dict[str, Any]]
        elements: list[dict[str, Any]]
        timeout: int | None = None

        def to_qua_config(self) -> DictQuaConfig:
            """Creates the Quantum Machines QUA config dictionary.

            Creates, an instance of a dictionary in the format that QuantumMachines expects the config dictionary to be.

            The values of following keys of the dictionary are parsed from QMMSettings:
                - controllers (OPX+ instruments)
                - octaves (Octave instruments)
                - mixers (up/down conversion settings if not using octaves)
                - elements (buses)

            The values of the rest keys are empty, meant to be updated in runtime.

            Returns:
                config: DictQuaConfig
            """
            elements, mixers = self._get_elements_and_mixers_config()
            controllers = self._get_controllers_config()
            octaves = self._get_octaves_config()

            return {
                "version": 1,
                "controllers": controllers,
                "elements": elements,
                "mixers": mixers,
                "waveforms": {},
                "integration_weights": {},
                "pulses": {},
                "digital_waveforms": {},
                "octaves": octaves,
            }

        def _get_controllers_config(self) -> dict[str, Any]:
            """Returns the controllers config dictionary.

            Returns:
                controllers: Dict[str, Any]
            """
            controllers: dict = {}
            for controller in self.controllers:
                controller_type = controller["type"] if "type" in controller else "opx1"
                if controller_type == "opx1":
                    controllers[controller["name"]] = {
                        "analog_outputs": {
                            output["port"]: {
                                "offset": output["offset"] if "offset" in output else 0.0,
                                "delay": output["delay"] if "delay" in output else 0.0,
                                "filter": (
                                    {
                                        "feedforward": (
                                            output["filter"]["feedforward"] if "feedforward" in output["filter"] else []
                                        ),
                                        "feedback": (
                                            output["filter"]["feedback"] if "feedback" in output["filter"] else []
                                        ),
                                    }
                                    if "filter" in output
                                    else {"feedforward": [], "feedback": []}
                                ),
                                "shareable": output["shareable"] if "shareable" in output else False,
                            }
                            for output in controller.get("analog_outputs", [])
                        },
                        "analog_inputs": {
                            input["port"]: {
                                "offset": input["offset"] if "offset" in input else 0.0,
                                "gain_db": input["gain"] if "gain" in input else 0.0,
                                "shareable": input["shareable"] if "shareable" in input else False,
                            }
                            for input in controller.get("analog_inputs", [])
                        },
                        "digital_outputs": {output["port"]: {} for output in controller.get("digital_outputs", [])},
                    }
                elif controller_type == "opx1000":
                    controllers[controller["name"]] = {
                        "fems": {
                            fem["fem"]: {
                                "type": fem["type"] if "type" in fem else "LF",
                                "analog_outputs": {
                                    output["port"]: {
                                        "offset": output["offset"] if "offset" in output else 0.0,
                                        "delay": output["delay"] if "delay" in output else 0.0,
                                        "output_mode": output["output_mode"] if "output_mode" in output else "direct",
                                        "sampling_rate": output["sampling_rate"] if "sampling_rate" in output else 1e9,
                                        "upsampling_mode": (
                                            output["upsampling_mode"]
                                            if "upsampling_mode" in output
                                            else (
                                                "pulse"
                                                if "output_mode" in output and output["output_mode"] == "amplified"
                                                else "mw"
                                            )
                                        ),
                                        "filter": (
                                            {
                                                "feedforward": (
                                                    output["filter"]["feedforward"]
                                                    if "feedforward" in output["filter"]
                                                    else []
                                                ),
                                                "feedback": (
                                                    output["filter"]["feedback"]
                                                    if "feedback" in output["filter"]
                                                    else []
                                                ),
                                            }
                                            if "filter" in output
                                            else {"feedforward": [], "feedback": []}
                                        ),
                                        "shareable": fem["shareable"] if "shareable" in fem else False,
                                    }
                                    for output in fem.get("analog_outputs", [])
                                },
                                "analog_inputs": {
                                    input["port"]: {
                                        "offset": input["offset"] if "offset" in input else 0.0,
                                        "gain_db": input["gain"] if "gain" in input else 0.0,
                                        "sampling_rate": input["sampling_rate"] if "sampling_rate" in input else 1e9,
                                        "shareable": fem["shareable"] if "shareable" in fem else False,
                                    }
                                    for input in fem.get("analog_inputs", [])
                                },
                                "digital_outputs": {output["port"]: {} for output in fem.get("digital_outputs", [])},
                            }
                            for fem in controller["fems"]
                        }
                    }
            return controllers

        def _get_octaves_config(self) -> dict[str, Any]:
            """Returns the octaves config dictionary.

            Returns:
                octaves: Dict[str, Any]
            """
            octaves: dict = {}
            for octave in self.octaves:
                octaves[octave["name"]] = {}
                octaves[octave["name"]]["RF_outputs"] = {}
                octaves[octave["name"]]["RF_inputs"] = {}
                for rf_output in octave.get("rf_outputs", []):
                    octaves[octave["name"]]["RF_outputs"][rf_output["port"]] = {
                        "LO_frequency": rf_output["lo_frequency"],  # Should be between 2 and 18 GHz.
                        "LO_source": "internal",
                        "gain": rf_output["gain"] if "gain" in rf_output else 0.0,
                        "output_mode": "always_on",
                        "input_attenuators": "OFF",  # can be: "OFF" / "ON". Default is "OFF".
                    }
                    if "i_connection" in rf_output:
                        octaves[octave["name"]]["RF_outputs"][rf_output["port"]]["I_connection"] = (
                            (
                                rf_output["i_connection"]["controller"],
                                rf_output["i_connection"]["fem"],
                                rf_output["i_connection"]["port"],
                            )
                            if "fem" in rf_output["i_connection"]
                            else (rf_output["i_connection"]["controller"], rf_output["i_connection"]["port"])
                        )
                    if "q_connection" in rf_output:
                        octaves[octave["name"]]["RF_outputs"][rf_output["port"]]["Q_connection"] = (
                            (
                                rf_output["q_connection"]["controller"],
                                rf_output["q_connection"]["fem"],
                                rf_output["q_connection"]["port"],
                            )
                            if "fem" in rf_output["q_connection"]
                            else (rf_output["q_connection"]["controller"], rf_output["q_connection"]["port"])
                        )
                for rf_input in octave.get("rf_inputs", []):
                    octaves[octave["name"]]["RF_inputs"][rf_input["port"]] = {
                        "RF_source": "RF_in",
                        "LO_frequency": rf_input["lo_frequency"],
                        "LO_source": "internal",  # can be: "internal" / "external". Default is "internal".
                        "IF_mode_I": "direct",  # can be: "direct" / "mixer" / "envelope" / "off". Default is "direct".
                        "IF_mode_Q": "direct",
                    }
                if "loopbacks" in octave:
                    octaves[octave["name"]]["loopbacks"] = [
                        (
                            (octave["name"], octave["loopbacks"]["Synth"]),
                            octave["loopbacks"]["Dmd"],
                        )
                    ]
                if "connectivity" in octave:
                    octaves[octave["name"]]["connectivity"] = (
                        (octave["connectivity"]["controller"], octave["connectivity"]["fem"])
                        if "fem" in octave
                        else octave["connectivity"]["controller"]
                    )
                else:
                    octaves[octave["name"]]["IF_outputs"] = {
                        "IF_out1": {
                            "port": (
                                (
                                    octave["if_outputs"][0]["controller"],
                                    octave["if_outputs"][0]["fem"],
                                    octave["if_outputs"][0]["port"],
                                )
                                if "fem" in octave["if_outputs"][0]
                                else (octave["if_outputs"][0]["controller"], octave["if_outputs"][0]["port"])
                            ),
                            "name": "out1",
                        },
                        "IF_out2": {
                            "port": (
                                (
                                    octave["if_outputs"][1]["controller"],
                                    octave["if_outputs"][1]["fem"],
                                    octave["if_outputs"][1]["port"],
                                )
                                if "fem" in octave["if_outputs"][1]
                                else (octave["if_outputs"][1]["controller"], octave["if_outputs"][1]["port"])
                            ),
                            "name": "out2",
                        },
                    }
            return octaves

        def _get_elements_and_mixers_config(self) -> tuple:
            """Returns the elements config dictionary.

            Returns:
                elements: Dict[str, Any]
            """
            elements = {}
            mixers = {}

            for element in self.elements:
                bus_name = str(element["identifier"])
                element_dict: dict = {"operations": {}}

                # Flux bus
                if "single_input" in element:
                    intermediate_frequency = (
                        int(element["intermediate_frequency"]) if "intermediate_frequency" in element else 0
                    )
                    element_dict["singleInput"] = {
                        "port": (
                            (
                                element["single_input"]["controller"],
                                element["single_input"]["fem"],
                                element["single_input"]["port"],
                            )
                            if "fem" in element["single_input"]
                            else (element["single_input"]["controller"], element["single_input"]["port"])
                        ),
                    }
                    element_dict["intermediate_frequency"] = intermediate_frequency
                # IQ bus
                elif "mix_inputs" in element:
                    mixer_name = f"mixer_{bus_name}"
                    lo_frequency = int(element["mix_inputs"]["lo_frequency"])
                    intermediate_frequency = int(element["intermediate_frequency"])
                    mixer_correction = (
                        element["mix_inputs"]["mixer_correction"]
                        if "mixer_correction" in element["mix_inputs"]
                        else [1.0, 0.0, 0.0, 1.0]
                    )
                    mixers[mixer_name] = [
                        {
                            "intermediate_frequency": intermediate_frequency,
                            "lo_frequency": lo_frequency,
                            "correction": mixer_correction,
                        }
                    ]
                    element_dict["mixInputs"] = {
                        key: (
                            (
                                element["mix_inputs"][key]["controller"],
                                element["mix_inputs"][key]["fem"],
                                element["mix_inputs"][key]["port"],
                            )
                            if "fem" in element["mix_inputs"][key]
                            else (element["mix_inputs"][key]["controller"], element["mix_inputs"][key]["port"])
                        )
                        for key in ["I", "Q"]
                    } | {"lo_frequency": lo_frequency, "mixer": mixer_name}
                    element_dict["intermediate_frequency"] = intermediate_frequency

                    # readout bus
                    if "outputs" in element:
                        element_dict["outputs"] = {
                            key: (
                                (
                                    element["outputs"][key]["controller"],
                                    element["outputs"][key]["fem"],
                                    element["outputs"][key]["port"],
                                )
                                if "fem" in element["outputs"][key]
                                else (element["outputs"][key]["controller"], element["outputs"][key]["port"])
                            )
                            for key in ["out1", "out2"]
                        }
                        element_dict["time_of_flight"] = element["time_of_flight"]
                        element_dict["smearing"] = element["smearing"]
                # RF with Octave
                elif "rf_inputs" in element:
                    element_dict["RF_inputs"] = {"port": (element["rf_inputs"]["octave"], element["rf_inputs"]["port"])}
                    element_dict["intermediate_frequency"] = int(element["intermediate_frequency"])

                    # readout bus
                    if "rf_outputs" in element:
                        element_dict["RF_outputs"] = {
                            "port": (element["rf_outputs"]["octave"], element["rf_outputs"]["port"]),
                        }
                        element_dict["time_of_flight"] = element["time_of_flight"]
                        element_dict["smearing"] = element["smearing"]

                # other settings
                if "digital_inputs" in element:
                    element_dict["digitalInputs"] = {
                        "switch": {
                            "port": (
                                (
                                    element["digital_inputs"]["controller"],
                                    element["digital_inputs"]["fem"],
                                    element["digital_inputs"]["port"],
                                )
                                if "fem" in element["digital_inputs"]
                                else (element["digital_inputs"]["controller"], element["digital_inputs"]["port"])
                            ),
                            "delay": element["digital_inputs"]["delay"],
                            "buffer": element["digital_inputs"]["buffer"],
                        }
                    }
                if "digital_outputs" in element:
                    element_dict["digitalOutputs"] = {
                        "port": (
                            (
                                element["digital_outputs"]["controller"],
                                element["digital_outputs"]["fem"],
                                element["digital_outputs"]["port"],
                            )
                            if "fem" in element["digital_outputs"]
                            else (element["digital_outputs"]["controller"], element["digital_outputs"]["port"])
                        ),
                    }

                elements[bus_name] = element_dict

            return elements, mixers

    settings: QuantumMachinesClusterSettings
    device: QMMDriver
    _qmm: QuantumMachinesManager
    _qm: QuantumMachine  # TODO: Change private QM API to public when implemented.
    _config: DictQuaConfig
    _octave_config: QmOctaveConfig | None = None
    _is_connected_to_qm: bool = False
    _config_created: bool = False
    _pending_set_intermediate_frequency: dict[str, float] = {}  # noqa: RUF012
    _compiled_program_cache: dict[str, str] = {}  # noqa: RUF012

    @property
    def config(self) -> DictQuaConfig:
        """Get the QUA config dictionary."""
        return self._config

    def is_awg(self) -> bool:
        """Returns True if instrument is an AWG."""
        return True

    def is_adc(self) -> bool:
        """Returns True if instrument is an ADC."""
        return True

    @check_device_initialized
    def initial_setup(self):
        """Sets initial instrument settings.

        Creates an instance of the Qililab Quantum Machines Manager, and sets the configuration dictionary.
        """
        if self.settings.octaves:
            self._octave_config = QmOctaveConfig()
            self._octave_config.set_calibration_db(os.getcwd())

        self._qmm = (
            QuantumMachinesManager(
                host=self.settings.address,
                cluster_name=self.settings.cluster,
                octave_calibration_db_path=self._octave_config._calibration_db_path,
            )
            if self._octave_config is not None
            else QuantumMachinesManager(
                host=self.settings.address,
                cluster_name=self.settings.cluster,
            )
        )
        self._config = self.settings.to_qua_config()
        self._config_created = True

    @check_device_initialized
    def turn_on(self):
        """Turns on the instrument."""
        if not self._is_connected_to_qm:
            self._qm = self._qmm.open_qm(config=self._config, close_other_machines=True)
            self._compiled_program_cache = {}
            self._is_connected_to_qm = True

            if self.settings.run_octave_calibration:
                self.run_octave_calibration()

    @check_device_initialized
    def reset(self):
        """Resets instrument settings."""

    @check_device_initialized
    def turn_off(self):
        """Turns off an instrument."""
        if self._is_connected_to_qm:
            self._qm.close()
            self._is_connected_to_qm = False

    def append_configuration(self, configuration: dict):
        """Update the `_config` dictionary by appending the configuration generated by compilation.

        Args:
            configuration (dict): Configuration dictionary to append to the existing configuration.

        Raises:
            ValueError: Raised if the `_config` dictionary does not exist.
        """
        if not self._config_created:
            raise ValueError("The QM `config` dictionary does not exist. Please run `initial_setup()` first.")

        merged_configuration = merge_dictionaries(dict(self._config), configuration)
        if self._config != merged_configuration:
            self._config = cast("DictQuaConfig", merged_configuration)
            # If we are already connected, reopen the connection with the new configuration
            if self._is_connected_to_qm:
                self._qm.close()
                self._qm = self._qmm.open_qm(config=self._config, close_other_machines=True)  # type: ignore[assignment]
                self._compiled_program_cache = {}

    def run_octave_calibration(self):
        """Run calibration procedure for the buses with octaves, if any."""
        elements = [element for element in self._config["elements"] if "RF_inputs" in self._config["elements"][element]]
        for element in elements:
            self._qm.calibrate_element(element)

    def get_controller_type_from_bus(self, bus: str) -> str | None:
        """Gets the OPX controller name of the bus used

        Args:
            bus (str): Alias of the bus

        Raises:
            AttributeError: Raised when given bus does not exist

        Returns:
            str | None: Alias of the controller, either opx1 or opx1000.
        """

        if "RF_inputs" in self._config["elements"][bus]:
            octave = self._config["elements"][bus]["RF_inputs"]["port"][0]
            controller_name = self._config["octaves"][octave]["connectivity"]
        elif "mixInputs" in self._config["elements"][bus]:
            controller_name = self._config["elements"][bus]["mixInputs"]["I"][0]
        elif "singleInput" in self._config["elements"][bus]:
            controller_name = self._config["elements"][bus]["singleInput"]["port"][0]

        for controller in self.settings.controllers:
            if controller["name"] == controller_name:
                return controller["type"] if "type" in controller else "opx1"
        raise AttributeError(f"Controller with bus {bus} does not exist")

    def get_controller_from_element(self, element: dict, key: str | None) -> tuple[str, int, int | None]:
        """Get controller name, port and FEM (if applicable) from element

        Args:
            element (dict): element of a bus
            key (str | None): Key for mix inputs, it can be I or Q.

        Returns:
            list: controller coordinates
        """
        if ("rf_inputs" in element or "mix_inputs" in element) and key not in ["I", "Q"]:
            raise ValueError(f"key value must be I or Q, {key} given")
        if "rf_inputs" in element:
            octave_name = element["rf_inputs"]["octave"]
            out_oct_port = element["rf_inputs"]["port"]

            octave = next((octave for octave in self.settings.octaves if octave["name"] == octave_name), None)
            octave_port = (
                next(
                    (octave_port for octave_port in octave["rf_outputs"] if octave_port["port"] == out_oct_port),
                    None,
                )
                if octave
                else None
            )

            connection = "i_connection" if key == "I" else "q_connection"
            if connection in octave_port:  # type: ignore[operator]
                con_name = octave_port[connection]["controller"]  # type: ignore[index]
                con_port = octave_port[connection]["port"]  # type: ignore[index]
                con_fem = octave_port[connection]["fem"] if "fem" in octave_port[connection] else None  # type: ignore[index]
            else:
                con_name = octave["connectivity"]["controller"]  # type: ignore[index]
                con_port = octave_port["port"] * 2 - 1 if key == "I" else octave_port["port"] * 2  # type: ignore[index]
                con_fem = None

        elif "mix_inputs" in element:
            con_name = element["mix_inputs"][key]["controller"]
            con_port = element["mix_inputs"][key]["port"]
            con_fem = element["mix_inputs"][key]["fem"] if "fem" in element["mix_inputs"][key] else None

        elif "single_input" in element:
            con_name = element["single_input"]["controller"]
            con_port = element["single_input"]["port"]
            con_fem = element["single_input"]["fem"] if "fem" in element["single_input"] else None

        return (con_name, con_port, con_fem)

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None) -> None:
        """Sets the parameter of the instrument into the cache (runtime dataclasses).

        And if connection to instruments is established, then to the instruments as well.

        Args:
            bus (str): The assossiated bus to change parameter.
            parameter (Parameter): The parameter to update.
            value (float | str | bool): The new value of the parameter.

        Raises:
            ValueError: Raised when passed bus is not found, or rf_inputs is not connected to an octave.
            ParameterNotFound: Raised if parameter does not exist.
        """
        bus = str(channel_id)
        element = next((element for element in self.settings.elements if element["identifier"] == bus), None)
        if element is None:
            raise ValueError(f"Bus {bus} was not found in {self.name} settings.")

        if parameter in [Parameter.LO_FREQUENCY, Parameter.GAIN]:
            if "rf_inputs" not in element:
                raise ValueError(
                    f"Trying to change parameter {parameter.name} in {self.name}, however bus {bus} is not connected to an octave."
                )
            octave_name = element["rf_inputs"]["octave"]
            out_port = element["rf_inputs"]["port"]
            in_port = element["rf_outputs"]["port"] if "rf_outputs" in element else None
            settings_octave = next(octave for octave in self.settings.octaves if octave["name"] == octave_name)
            settings_octave_rf_output = next(
                rf_output for rf_output in settings_octave["rf_outputs"] if rf_output["port"] == out_port
            )

        # Now we will set the parameter in 3 places:
        # 1) In the settings runtime dataclass (always).
        # 2) If created: In the `_config`` dictionary.
        # 3) If connected: In the instrument itself.
        if parameter == Parameter.LO_FREQUENCY:
            lo_frequency = float(value)
            settings_octave_rf_output["lo_frequency"] = lo_frequency
            if self._config_created:
                self._config["octaves"][octave_name]["RF_outputs"][out_port]["LO_frequency"] = lo_frequency
            if self._is_connected_to_qm:
                self._qm.octave.set_lo_frequency(element=bus, lo_frequency=lo_frequency)
                self._qm.calibrate_element(bus)
            if in_port is not None:
                settings_octave_rf_input = next(
                    rf_input for rf_input in settings_octave["rf_inputs"] if rf_input["port"] == in_port
                )
                settings_octave_rf_input["lo_frequency"] = lo_frequency
                if self._config_created:
                    self._config["octaves"][octave_name]["RF_inputs"][in_port]["LO_frequency"] = lo_frequency
            return

        if parameter == Parameter.GAIN:
            gain_in_db = float(value)
            settings_octave_rf_output["gain"] = gain_in_db
            if self._config_created:
                self._config["octaves"][octave_name]["RF_outputs"][out_port]["gain"] = gain_in_db
            if self._is_connected_to_qm:
                self._qm.octave.set_rf_output_gain(element=bus, gain_in_db=gain_in_db)
            return

        if parameter == Parameter.IF:
            intermediate_frequency = float(value)
            element["intermediate_frequency"] = intermediate_frequency
            if self._config_created:
                self._config["elements"][bus]["intermediate_frequency"] = intermediate_frequency
                if f"mixer_{bus}" in self._config["mixers"]:
                    self._config["mixers"][f"mixer_{bus}"][0]["intermediate_frequency"] = intermediate_frequency
            if self._is_connected_to_qm:
                controller_type = self.get_controller_type_from_bus(bus)
                if controller_type == "opx1":
                    self._qm.set_intermediate_frequency(element=bus, freq=intermediate_frequency)
                if controller_type == "opx1000":
                    self._pending_set_intermediate_frequency[bus] = intermediate_frequency
            return

        if parameter == Parameter.THRESHOLD_ROTATION:
            threshold_rotation = float(value)
            element["threshold_rotation"] = threshold_rotation
            return

        if parameter == Parameter.THRESHOLD:
            threshold = float(value)
            element["threshold"] = threshold
            return

        if parameter == Parameter.DC_OFFSET:
            con_name, con_port, con_fem = self.get_controller_from_element(element=element, key=None)
            dc_offset = float(value)
            settings_controllers = next(
                controller for controller in self.settings.controllers if controller["name"] == con_name
            )
            if con_fem is None:
                settings_offset = next(
                    analog_output
                    for analog_output in settings_controllers["analog_outputs"]
                    if analog_output["port"] == con_port
                )
            else:
                settings_fem = next(fem for fem in settings_controllers["fems"] if fem["fem"] == con_fem)
                settings_offset = next(
                    analog_output
                    for analog_output in settings_fem["analog_outputs"]
                    if analog_output["port"] == con_port
                )
            settings_offset["offset"] = dc_offset
            if self._config_created:
                if con_fem is None:
                    self._config["controllers"][con_name]["analog_outputs"][con_port]["offset"] = dc_offset  # type: ignore[typeddict-item]
                else:
                    self._config["controllers"][con_name]["fems"][con_fem]["analog_outputs"][con_port][  # type: ignore[typeddict-item, index]
                        "offset"  # type: ignore[typeddict-unknown-key]
                    ] = dc_offset
            if self._is_connected_to_qm:
                self._qm.set_output_dc_offset_by_element(element=bus, input="single", offset=dc_offset)
            return

        if parameter in [Parameter.OFFSET_I, Parameter.OFFSET_Q]:
            key = "I" if parameter == Parameter.OFFSET_I else "Q"
            con_name, con_port, con_fem = self.get_controller_from_element(element=element, key=key)
            input_offset = float(value)
            settings_controllers = next(
                controller for controller in self.settings.controllers if controller["name"] == con_name
            )
            if con_fem is None:
                settings_offset = next(
                    analog_output
                    for analog_output in settings_controllers["analog_outputs"]
                    if analog_output["port"] == con_port
                )
            else:
                settings_fem = next(fem for fem in settings_controllers["fems"] if fem["fem"] == con_fem)
                settings_offset = next(
                    analog_output
                    for analog_output in settings_fem["analog_outputs"]
                    if analog_output["port"] == con_port
                )
            settings_offset["offset"] = input_offset
            if self._config_created:
                if con_fem is None:
                    self._config["controllers"][con_name]["analog_outputs"][con_port]["offset"] = input_offset  # type: ignore[typeddict-item]
                else:
                    self._config["controllers"][con_name]["fems"][con_fem]["analog_outputs"][con_port][  # type: ignore[typeddict-item, index]
                        "offset"  # type: ignore[typeddict-unknown-key]
                    ] = input_offset
            if self._is_connected_to_qm:
                self._qm.set_output_dc_offset_by_element(element=bus, input=key, offset=input_offset)
            return

        if parameter in [Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2]:
            output = "out1" if parameter == Parameter.OFFSET_OUT1 else "out2"
            out_value = 1 if output == "out1" else 2
            con_name, _, con_fem = self.get_controller_from_element(element=element, key="I")
            output_offset = float(value)
            settings_controllers = next(
                controller for controller in self.settings.controllers if controller["name"] == con_name
            )
            if con_fem is None:
                settings_offset = next(
                    analog_output
                    for analog_output in settings_controllers["analog_inputs"]
                    if analog_output["port"] == out_value
                )
            else:
                settings_fem = next(fem for fem in settings_controllers["fems"] if fem["fem"] == con_fem)
                settings_offset = next(
                    analog_output
                    for analog_output in settings_fem["analog_inputs"]
                    if analog_output["port"] == out_value
                )
            settings_offset["offset"] = output_offset
            if self._config_created:
                if con_fem is None:
                    self._config["controllers"][con_name]["analog_inputs"][out_value]["offset"] = output_offset  # type: ignore[typeddict-item]
                else:
                    self._config["controllers"][con_name]["fems"][con_fem]["analog_inputs"][out_value][  # type: ignore[typeddict-item, index]
                        "offset"  # type: ignore[typeddict-unknown-key]
                    ] = output_offset
            if self._is_connected_to_qm:
                self._qm.set_input_dc_offset_by_element(element=bus, output=output, offset=output_offset)
            return

        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None) -> ParameterValue:
        """Gets the value of a parameter.

        Args:
            bus (str): The associated bus of the parameter.
            parameter (Parameter): The parameter to get value.

        Returns:
            float | int | bool | tuple: The value of the parameter.

        Raises:
            ParameterNotFound: Raised if parameter does not exist.
        """
        # Just in case, read from the `settings`, even though in theory the config should always be synch:
        bus = str(channel_id)
        settings_config_dict = self.settings.to_qua_config()
        config_keys = settings_config_dict["elements"][bus]
        element = next((element for element in self.settings.elements if element["identifier"] == bus), None)

        if parameter == Parameter.LO_FREQUENCY:
            if "mixInputs" in config_keys:
                return settings_config_dict["elements"][bus]["mixInputs"]["lo_frequency"]
            if "RF_inputs" in config_keys:
                port = settings_config_dict["elements"][bus]["RF_inputs"]["port"]
                return settings_config_dict["octaves"][port[0]]["RF_outputs"][port[1]]["LO_frequency"]

        if parameter == Parameter.IF:
            if "intermediate_frequency" in config_keys:
                return settings_config_dict["elements"][bus]["intermediate_frequency"]

        if parameter == Parameter.GAIN:
            if "mixInputs" in config_keys and "outputs" in config_keys:
                port_i = settings_config_dict["elements"][bus]["outputs"]["out1"]
                # port_q = settings_config_dict["elements"][bus]["outputs"]["out2"]
                return settings_config_dict["controllers"][port_i[0]]["analog_inputs"][port_i[1]]["gain_db"]  # type: ignore[typeddict-item]
                # return (
                #     settings_config_dict["controllers"][port_i[0]]["analog_inputs"][port_i[1]]["gain_db"],  # type: ignore[typeddict-item]
                #     settings_config_dict["controllers"][port_q[0]]["analog_inputs"][port_q[1]]["gain_db"],  # type: ignore[typeddict-item]
                # )
            if "RF_inputs" in config_keys:
                port = settings_config_dict["elements"][bus]["RF_inputs"]["port"]
                return settings_config_dict["octaves"][port[0]]["RF_outputs"][port[1]]["gain"]

        if parameter == Parameter.TIME_OF_FLIGHT:
            if "time_of_flight" in config_keys:
                return settings_config_dict["elements"][bus]["time_of_flight"]

        if parameter == Parameter.SMEARING:
            if "smearing" in config_keys:
                return settings_config_dict["elements"][bus]["smearing"]

        if parameter in [Parameter.THRESHOLD_ROTATION, Parameter.THRESHOLD]:
            element = next((element for element in self.settings.elements if element["identifier"] == bus), None)
            if parameter == Parameter.THRESHOLD_ROTATION:
                return element.get("threshold_rotation", None)  # type: ignore
            if parameter == Parameter.THRESHOLD:
                return element.get("threshold", None)  # type: ignore

        if parameter == Parameter.DC_OFFSET:
            con_name, con_port, con_fem = self.get_controller_from_element(element=element, key=None)  # type: ignore[arg-type]
            if con_fem is None:
                return settings_config_dict["controllers"][con_name]["analog_outputs"][con_port]["offset"]  # type: ignore[typeddict-item]
            return settings_config_dict["controllers"][con_name]["fems"][con_fem]["analog_outputs"][con_port]["offset"]  # type: ignore[typeddict-item, index]

        if parameter in [Parameter.OFFSET_I, Parameter.OFFSET_Q]:
            key = "I" if parameter in Parameter.OFFSET_I else "Q"
            con_name, con_port, con_fem = self.get_controller_from_element(element=element, key=key)  # type: ignore[arg-type]
            if con_fem is None:
                return settings_config_dict["controllers"][con_name]["analog_outputs"][con_port]["offset"]  # type: ignore[typeddict-item]
            return settings_config_dict["controllers"][con_name]["fems"][con_fem]["analog_outputs"][con_port]["offset"]  # type: ignore[typeddict-item, index]

        if parameter in [Parameter.OFFSET_OUT1, Parameter.OFFSET_OUT2]:
            output = "out1" if parameter in Parameter.OFFSET_OUT1 else "out2"
            out_value = 1 if output == "out1" else 2
            con_name, _, con_fem = self.get_controller_from_element(element=element, key="I")  # type: ignore[arg-type]
            if con_fem is None:
                return settings_config_dict["controllers"][con_name]["analog_inputs"][out_value]["offset"]  # type: ignore[typeddict-item]
            return settings_config_dict["controllers"][con_name]["fems"][con_fem]["analog_inputs"][out_value]["offset"]  # type: ignore[typeddict-item, index]

        raise ParameterNotFound(self, parameter)

    def compile(self, program: Program) -> str:
        """Compiles and stores a given QUA program on the Quantum Machines instance,
        and returns a unique identifier associated with the compiled program.

        The method first generates a hash for the input QUA program. If this hash is not already present in
        the cache, it proceeds to compile the program using QM's compile method and stores the result in the cache
        indexed by the generated hash. This caching mechanism prevents recompiling the same program multiple times,
        thus optimizing performance.

        Args:
            program (Program): The QUA program to be compiled.

        Returns:
            str: A unique identifier (hash) for the compiled QUA program. This identifier can be used to retrieve the compiled program from the cache, or run it  with `run_compiled_program` method.
        """
        qua_program_hash = hash_qua_program(program=program)
        if qua_program_hash not in self._compiled_program_cache:
            self._compiled_program_cache[qua_program_hash] = self._qm.compile(program=program)
        return self._compiled_program_cache[qua_program_hash]

    def run_compiled_program(self, compiled_program_id: str) -> QmJob | JobApi:
        """Executes a previously compiled QUA program identified by its unique compiled program ID.

        This method submits the compiled program to the Quantum Machines (QM) execution queue and waits for
        its execution to complete. The execution of the program is managed by the QM's queuing system,
        which schedules and runs the job on the quantum hardware or simulator as per availability and queue status.

        Args:
            compiled_program_id (str): The unique identifier of the compiled QUA program to be executed. This ID should correspond to a program that has already been compiled and is present in the program cache.

        Returns:
            RunningQmJob: An object representing the running job. This object provides methods and properties to check the status of the job, retrieve results upon completion, and manage or investigate the job's execution.
        """
        # TODO: qm.queue.add_compiled() -> qm.add_compiled()
        pending_job = self._qm.queue.add_compiled(compiled_program_id)

        # TODO: job.wait_for_execution() is deprecated and will be removed in the future. Please use job.wait_until("Running") instead.
        job = pending_job.wait_for_execution()  # type: ignore[return-value, union-attr]
        if self._pending_set_intermediate_frequency:
            for bus, intermediate_frequency in self._pending_set_intermediate_frequency.items():
                job.set_intermediate_frequency(element=bus, freq=intermediate_frequency)  # type: ignore[union-attr]
                self._qm.calibrate_element(bus)
            self._pending_set_intermediate_frequency = {}

        return job

    def run(self, program: Program) -> RunningQmJob:
        """Runs the QUA Program.

        Creates, for every execution, an instance of a QuantumMachine class and executes the QUA program on it.
        After execution, the job generated by the Quantum Machines Manager is returned.

        Args:
            program (Program): QUA Program to be run on Quantum Machines instruments.

        Returns:
            job: Quantum Machines job.
        """

        return self._qm.execute(program)

    def get_acquisitions(self, job: QmJob | JobApi) -> dict[str, np.ndarray]:
        """Fetches the results from the execution of a QUA Program.

        Once the results have been fetched, they are returned wrapped in a QuantumMachinesResult instance.

        Args:
            job (QmJob): Job that provides the result handles.

        Returns:
            QuantumMachinesResult: Quantum Machines result instance.
        """
        result_handles_fetchers = job.result_handles
        result_handles_fetchers.wait_for_all_values(timeout=self.settings.timeout)
        results = {
            name: handle.fetch_all(flat_struct=True) for name, handle in job.result_handles if handle is not None
        }
        return {name: data for name, data in results.items() if data is not None}

    def simulate(self, program: Program) -> RunningQmJob:
        """Simulates the QUA Program.

        Creates, for every simulation, an instance of a QuantumMachine class and executes the QUA program on it.
        It returns the running job instance.

        Args:
            program (Program): QUA Program to be simulated on Quantum Machines instruments.

        Returns:
            job: Quantum Machines job.
        """
        return self._qm.simulate(program, SimulationConfig(40_000))



""" Test Results """
import os

import numpy as np
import pytest

from qililab.result.qprogram.qblox_measurement_result import QbloxMeasurementResult
from qililab.utils.serialization import deserialize, deserialize_from, serialize, serialize_to


@pytest.fixture(name="raw_measurement_data")
def fixture_raw_measurement_data() -> dict:
    """Dictionary of raw measurement data as returned from QRM instruments."""
    return {"bins": {"integration": {"path0": [1, 2, 3,], "path1": [4, 5, 6]}, "threshold": [0.1, 0.2, 0.3]}}

@pytest.fixture(name="raw_measurement_data_nested_loops")
def fixture_raw_measurement_data_nested() -> dict:
    """Dictionary of raw measurement data as returned from QRM instruments."""
    return {"bins": {"integration": {"path0": [1, 2, 3, 4, 5, 6], "path1": [7, 8, 9, 10, 11, 12]}, "threshold": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6]}}

@pytest.fixture(name="qblox_measurement_result")
def fixture_qblox_result_noscope(raw_measurement_data: dict):
    """Instance of QbloxMeasurementResult."""
    return QbloxMeasurementResult(bus="readout", raw_measurement_data=raw_measurement_data)

@pytest.fixture(name="qblox_measurement_result_shape")
def fixture_qblox_result_with_shape(raw_measurement_data: dict):
    """Instance of QbloxMeasurementResult."""
    return QbloxMeasurementResult(bus="readout", raw_measurement_data=raw_measurement_data, shape=(1, 3))

@pytest.fixture(name="qblox_measurement_result_nested_loops")
def fixture_qblox_result_with_nested_loop(raw_measurement_data_nested_loops: dict):
    """Instance of QbloxMeasurementResult."""
    return QbloxMeasurementResult(bus="readout", raw_measurement_data=raw_measurement_data_nested_loops, shape=(2, 3))


class TestsQbloxQProgramMeasurementResult:
    """Test `QbloxQProgramResults` functionalities."""

    def test_init(self, qblox_measurement_result: QbloxMeasurementResult):
        """Test the instantiation of QbloxMeasurementResult."""
        assert isinstance(qblox_measurement_result, QbloxMeasurementResult)
        assert isinstance(qblox_measurement_result.raw_measurement_data, dict)
        assert "bins" in qblox_measurement_result.raw_measurement_data
        assert "integration" in qblox_measurement_result.raw_measurement_data["bins"]
        assert "path0" in qblox_measurement_result.raw_measurement_data["bins"]["integration"]
        assert "path1" in qblox_measurement_result.raw_measurement_data["bins"]["integration"]

    def test_array_property(self, raw_measurement_data: dict, qblox_measurement_result: QbloxMeasurementResult):
        """Test the array property returns the correct data."""
        path0 = raw_measurement_data["bins"]["integration"]["path0"]
        path1 = raw_measurement_data["bins"]["integration"]["path1"]
        expected_array = np.array([path0, path1])

        assert np.allclose(qblox_measurement_result.array, expected_array)

    def test_array_property_with_shape(self, raw_measurement_data: dict, qblox_measurement_result_shape: QbloxMeasurementResult):
        """Test the array property returns the correct data."""
        path0 = raw_measurement_data["bins"]["integration"]["path0"]
        path1 = raw_measurement_data["bins"]["integration"]["path1"]
        expected_array = np.array([[path0], [path1]])

        assert np.allclose(qblox_measurement_result_shape.array, expected_array)
        assert qblox_measurement_result_shape.array.shape == expected_array.shape

    def test_array_property_nested_loops(self, raw_measurement_data_nested_loops: dict, qblox_measurement_result_nested_loops: QbloxMeasurementResult):
        """Test the array property returns the correct data."""
        path0 = raw_measurement_data_nested_loops["bins"]["integration"]["path0"]
        path1 = raw_measurement_data_nested_loops["bins"]["integration"]["path1"]
        expected_array = np.array([path0, path1]).reshape((2, 2, 3))

        assert np.allclose(qblox_measurement_result_nested_loops.array, expected_array)
        assert qblox_measurement_result_nested_loops.array.shape == expected_array.shape

    def test_serialization_deserialization(self, qblox_measurement_result: QbloxMeasurementResult):
        """Test serialization and deserialization works."""
        serialized = serialize(qblox_measurement_result)
        deserialized_qblox_measurement_result = deserialize(serialized, QbloxMeasurementResult)

        assert isinstance(deserialized_qblox_measurement_result, QbloxMeasurementResult)

        serialize_to(qblox_measurement_result, file="qblox_measurement_result.yml")
        deserialized_qblox_measurement_result = deserialize_from("qblox_measurement_result.yml", QbloxMeasurementResult)

        assert isinstance(deserialized_qblox_measurement_result, QbloxMeasurementResult)

        os.remove("qblox_measurement_result.yml")

    def test_threshold(self, qblox_measurement_result: QbloxMeasurementResult):
        """Test the thresholded data as an np.ndarray"""
        thresholded_data = qblox_measurement_result.threshold
        expected_thresholds = np.array(qblox_measurement_result.raw_measurement_data["bins"]["threshold"])

        assert isinstance(thresholded_data, np.ndarray)
        assert np.all(thresholded_data == expected_thresholds)
        assert thresholded_data.shape == expected_thresholds.shape

    def test_threshold_nested_loops(self, qblox_measurement_result_nested_loops: QbloxMeasurementResult):
        """Test the thresholded data as an np.ndarray"""
        thresholded_data = qblox_measurement_result_nested_loops.threshold
        expected_thresholds = np.array(qblox_measurement_result_nested_loops.raw_measurement_data["bins"]["threshold"])
        expected_thresholds = expected_thresholds.reshape((1, 2, 3))

        assert isinstance(thresholded_data, np.ndarray)
        assert np.all(thresholded_data == expected_thresholds)
        assert thresholded_data.shape == expected_thresholds.shape



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gaussian pulse shape."""

from dataclasses import dataclass

import numpy as np

from qililab.pulse.pulse_shape.pulse_shape import PulseShape
from qililab.typings import PulseShapeName
from qililab.utils import Factory


@Factory.register
@dataclass(frozen=True, eq=True)
class Gaussian(PulseShape):
    r"""Standard centered Gaussian pulse shape, symmetrically spanning for ``num_sigmas`` over the pulse duration.

    The normal distribution's parameters :math:`\mu` (mean) and :math:`\sigma` (standard deviation) will be therefore
    defined by :math:`\mu =` ``duration`` :math:`/ 2` and :math:`\sigma =` ``duration`` :math:`/` ``num_sigmas``:

    .. math::

        Gaussian(x) = A * exp(-0.5 * (x - \mu)^2 / \sigma^2)

    You can think of it, as if an infinite expanding gaussian, is symmetrically cut in the given ``num_sigmas``, then that cut is
    expanded to occupy all the pulse duration, and then is shifted down so that it starts at 0.

    Examples:

        To get the envelope of a gaussian, with ``num_sigmas`` equal to ``X``, you need to do:

        .. code-block:: python

            from qililab.pulse.pulse_shape import Gaussian

            gaussian_envelope = Gaussian(num_sigmas=X).envelope(duration=50, amplitude=1)

        which for ``X`` being ``1``, ``4``, ``6`` or ``8``, look respectively like:

        .. image:: /classes_images/gaussians.png
            :width: 800
            :align: center

        This comes from the following steps, parting from a "full" gaussian (big ``num_sigma``) [blue]:
            1. you cut a full gaussian to the given ``num_sigmas``, for example ``1`` [small purple].
            2. in reality this cut gaussian, actually extends all the duration [big purple],
            3. you shift/distortion down the cut gaussian, so it starts at 0 height [red].

        .. image:: /classes_images/gaussian_explanation.png
            :width: 400
            :align: center

    Args:
        num_sigmas (float): Sigma number of the gaussian pulse shape. Defines the width of the gaussian pulse.
    """

    name = PulseShapeName.GAUSSIAN  # noqa: RUF045
    num_sigmas: float  #: Sigma number of the gaussian pulse shape.

    def envelope(self, duration: int, amplitude: float, resolution: float = 1.0) -> np.ndarray:
        """Gaussian envelope centered with respect to the pulse.

        The gaussian is symmetrically cut in the given ``num_sigmas``, meaning that it starts and ends at that sigma width.

        And then to avoid introducing noise at time 0, the full gaussian is shifted down so that it starts at 0.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """
        sigma = duration / self.num_sigmas
        time = np.arange(duration / resolution) * resolution
        mu_ = duration / 2

        gaussian = amplitude * np.exp(-0.5 * (time - mu_) ** 2 / sigma**2)
        norm = np.amax(np.abs(np.real(gaussian)))

        gaussian = gaussian - gaussian[0]  # Shift to avoid introducing noise at time 0
        corr_norm = np.amax(np.abs(np.real(gaussian)))

        return gaussian * norm / corr_norm if corr_norm != 0 else gaussian



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass

from qililab.utils.asdict_factory import dict_factory


@dataclass
class QbloxSequencer:
    identifier: int
    outputs: list[int]  # [3, 2] means I path is connected to output 3 and Q path is connected to output 2
    intermediate_frequency: float
    gain_imbalance: float | None
    phase_imbalance: float | None
    hardware_modulation: bool
    gain_i: float
    gain_q: float
    offset_i: float
    offset_q: float

    def to_dict(self):
        """Return a dict representation of an AWG Sequencer."""
        return asdict(self, dict_factory=dict_factory) | {"outputs": self.outputs}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""IQPair dataclass."""

from __future__ import annotations

from dataclasses import dataclass

from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.variable import Domain
from qililab.waveforms.drag_correction import DragCorrection
from qililab.waveforms.gaussian import Gaussian
from qililab.waveforms.waveform import Waveform
from qililab.yaml import yaml


@dataclass
@yaml.register_class
class IQPair:
    """IQPair dataclass, containing the 'in-phase' (I) and 'quadrature' (Q) parts of a signal."""

    I: Waveform
    Q: Waveform

    def __post_init__(self):
        if not isinstance(self.I, Waveform) or not isinstance(self.Q, Waveform):
            raise TypeError("Waveform inside IQPair must have Waveform type.")

        if self.I.get_duration() != self.Q.get_duration():
            raise ValueError("Waveforms of an IQ pair must have the same duration.")

    def get_duration(self) -> int:
        """Get the duration of the waveforms

        Returns:
            int: The duration of the waveforms.
        """
        return self.I.get_duration()

    @requires_domain("amplitude", Domain.Voltage)
    @requires_domain("duration", Domain.Time)
    @requires_domain("num_sigmas", Domain.Scalar)
    @requires_domain("drag_coefficient", Domain.Scalar)
    @staticmethod
    def DRAG(amplitude: float, duration: int, num_sigmas: float, drag_coefficient: float) -> IQPair:
        """Create a DRAG pulse. This is an IQ pair where the I channel corresponds to the gaussian wave and the Q is the drag correction, which corresponds to the derivative of the I channel times a ``drag_coefficient``.

        Args:
            amplitude (float): Maximum amplitude of the pulse.
            duration (int): Duration of the pulse (ns).
            num_sigmas (float): Sigma number of the gaussian pulse shape. Defines the width of the gaussian pulse.
            drag_coefficient (float): Drag coefficient that gives the DRAG its imaginary components.
        """
        waveform_i = Gaussian(amplitude=amplitude, duration=duration, num_sigmas=num_sigmas)
        waveform_q = DragCorrection(drag_coefficient=drag_coefficient, waveform=waveform_i)

        return IQPair(I=waveform_i, Q=waveform_q)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Chained waveform."""

import numpy as np

from qililab.yaml import yaml

from .waveform import Waveform


@yaml.register_class
class Chained(Waveform):
    """Represents a composite waveform consisting of multiple waveforms played in a chained sequence.

    This class allows combining multiple waveforms (such as Ramp, Square, etc.) into a single waveform by playing them one after the other.

    Args:
        waveforms (list[Waveform]): A list of individual waveform objects that will be chained together in sequence.

    Examples:
        The following example creates a Chained waveform composed of a Ramp, a Square, and another Ramp.

        .. code-block:: python

            import qililab as ql

            # Create the waveform
            chained_wf = ql.Chained(waveforms=[
                ql.Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=50)
                ql.Square(amplitude=1.0, duration=100)
                ql.Ramp(from_amplitude=1.0, to_amplitude=0.0, duration=50)
            ])

            # Get waveform's envelope
            envelope = chained_wf.envelope()

            # Get waveform's duration
            envelope = chained_wf.get_duration()
    """

    def __init__(self, waveforms: list[Waveform]) -> None:
        """
        Initialize the Chained waveform.

        Args:
            waveforms (list[Waveform]): A list of waveform instances to be chained together in sequence.
        """
        super().__init__()
        self.waveforms = waveforms

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Retrieve the envelope of the chained waveform.

        The envelope is obtained by concatenating the envelopes of each individual waveform in the sequence, with the specified resolution.

        Args:
            resolution (int, optional): The resolution of the waveform (number of time steps per unit of duration). Defaults to 1.

        Returns:
            np.ndarray: An array representing the amplitude of the waveform at each time step.
        """
        return np.concatenate([waveform.envelope(resolution=resolution) for waveform in self.waveforms])

    def get_duration(self) -> int:
        """Retrieve the total duration of the chained waveform.

        The total duration is calculated as the sum of the durations of each individual waveform in the sequence.

        Returns:
            int: The total duration of the waveform in nanoseconds.
        """
        return sum(waveform.get_duration() for waveform in self.waveforms)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Waveform protocol class."""

from abc import ABC, abstractmethod

import numpy as np


class Waveform(ABC):
    """Waveforms describes the pulses envelope's shapes. ``Waveform`` is their abstract base class.

    Every child of this interface needs to contain an `envelope` method.

    The `envelope` method will create the corresponding array of each shape.

    Derived: :class:`Arbitrary`,  :class:`Square`, :class:`Gaussian` and :class:`DragCorrection`.
    """

    @abstractmethod
    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Returns the pulse height for each time step.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """

    def get_duration(self) -> int:
        """Get the duration of the waveform.

        Returns:
            int: The duration of the waveform in ns.
        """
        return len(self.envelope())



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QDevil QDAC-II Instrument"""

from dataclasses import dataclass

import numpy as np

from qililab.instruments import InstrumentFactory, ParameterNotFound, check_device_initialized, log_set_parameter
from qililab.instruments.voltage_source import VoltageSource
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings import QDevilQDac2 as QDevilQDac2Driver
from qililab.waveforms import Waveform


@InstrumentFactory.register
class QDevilQDac2(VoltageSource):
    """QDevil QDAC-II Intrument

    Args:
        name (InstrumentName): Name of the instrument
        device (QDevilQDac2): Instance of the QCoDes QDac2 class.
        settings (QDevilQDac2Settings): Settings of the instrument.
    """

    _MIN_RAMPING_RATE: float = 0.01
    _MAX_RAMPING_RATE: float = 2e7
    name = InstrumentName.QDEVIL_QDAC2

    @dataclass
    class QDevilQDac2Settings(VoltageSource.VoltageSourceSettings):
        """Contains the settings of a specific signal generator."""

        low_pass_filter: list[str]
        mode: str = "offset"

    settings: QDevilQDac2Settings
    device: QDevilQDac2Driver
    _cache: dict[int | str, bool] = {}  # noqa: RUF012

    @property
    def low_pass_filter(self):
        """QDAC-II `low_pass_filter` property.

        Returns:
            list[str]: A list of the low pass filter setting of each DACs.
        """
        return self.settings.low_pass_filter

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set parameter to the corresponding value for an instrument's channel.

        Args:
            parameter (Parameter): Name of the parameter to be updated
            value (float | str | bool): New value of the parameter
            channel_id (int | None): Channel identifier
        """
        self._validate_channel(channel_id=channel_id)

        channel = self.device.channel(channel_id) if self.is_device_active() else None

        index = self.dacs.index(channel_id)
        if parameter == Parameter.VOLTAGE:
            voltage = float(value)
            self.settings.voltage[index] = voltage
            if self.is_device_active():
                channel.dc_constant_V(voltage)  # type: ignore[union-attr]
            return
        if parameter == Parameter.SPAN:
            span = str(value)
            self.settings.span[index] = span
            if self.is_device_active():
                channel.output_range(span)  # type: ignore[union-attr]
            return
        if parameter == Parameter.RAMPING_ENABLED:
            ramping_enabled = bool(value)
            self.settings.ramping_enabled[index] = ramping_enabled
            if self.is_device_active():
                if ramping_enabled:
                    if (
                        self.ramp_rate[index] < QDevilQDac2._MIN_RAMPING_RATE
                        or self.ramp_rate[index] > QDevilQDac2._MAX_RAMPING_RATE
                    ):
                        raise ValueError(
                            f"The ramp rate is out of range on channel {channel_id}. It should be between 0.01 V/s and 2e7 V/s."
                        )
                    channel.dc_slew_rate_V_per_s(self.ramp_rate[index])  # type: ignore[union-attr]
                else:
                    channel.dc_slew_rate_V_per_s(2e7)  # type: ignore[union-attr]
            return
        if parameter == Parameter.RAMPING_RATE:
            ramping_rate = float(value)
            self.settings.ramp_rate[index] = ramping_rate
            ramping_enabled = self.ramping_enabled[index]
            if ramping_enabled and self.is_device_active():
                if ramping_rate < QDevilQDac2._MIN_RAMPING_RATE or ramping_rate > QDevilQDac2._MAX_RAMPING_RATE:
                    raise ValueError(
                        f"The ramp rate is out of range on channel {channel_id}. It should be between 0.01 V/s and 2e7 V/s."
                    )
                channel.dc_slew_rate_V_per_s(ramping_rate)  # type: ignore[union-attr]
            return
        if parameter == Parameter.LOW_PASS_FILTER:
            low_pass_filter = str(value)
            self.settings.low_pass_filter[index] = low_pass_filter
            if self.is_device_active():
                channel.output_filter(low_pass_filter)  # type: ignore[union-attr]
            return
        raise ParameterNotFound(self, parameter)

    def get_dac(self, channel_id: ChannelID):
        """Get specific DAC from QDAC.

        Args:
            channel_id (ChannelID): channel id of the dac
        """
        return self.device.channel(channel_id)

    def upload_waveform(self, waveform: Waveform, channel_id: ChannelID):
        """Uploads a waveform to the instrument and saves it to _cache.
        IMPORTANT: note that the waveform resolution is not to the ns, it is acutally around 1_micro_second.

        Args:
            waveform (Waveform): Waveform to upload
            channel_id (ChannelID): channel id of the qdac

        Raises:
            ValueError: if a waveform is already allocated
        """
        envelope = waveform.envelope()
        values = list(envelope)  # TODO: does np array work?
        if channel_id in self._cache:
            raise ValueError(
                f"Device {self.name} already has a waveform allocated to channel {channel_id}. Clear the cache before allocating a new waveform"
            )
        # check that waveform entries are multiple of 2, check that amplitudes are within [-1,1] range
        if len(envelope) % 2 != 0:
            raise ValueError("Waveform entries must be even.")
        if np.max(np.abs(envelope)) >= 1:
            raise ValueError("Waveform amplitudes must be within [-1,1] range.")
        trace = self.device.allocate_trace(channel_id, len(values))
        trace.waveform(values)
        self._cache[channel_id] = True

    def play(self, channel_id: ChannelID | None = None, clear_after: bool = True):
        """Plays a waveform for a given channel id. If no channel id is given, plays all waveforms stored in the cache.

        Args:
            channel_id (ChannelID, optional): Channel id to play a waveform through. Defaults to None.
            clear_after (bool): If True, clears cache. Defaults to True.
        """
        if channel_id is None:
            for dac in self.dacs:
                awg_context = self.get_dac(dac).arbitrary_wave(dac)
            self.device.start_all()
        else:
            awg_context = self.get_dac(channel_id).arbitrary_wave(channel_id)
            awg_context.start()
        if clear_after:
            self.clear_cache()

        # TODO: catch errors raised at self.device.errors()

    def clear_cache(self):
        """Clears the cache of the instrument"""
        self.device.remove_traces()  # TODO: this method should be run at initial setup if instrument is in awg mode
        self._cache = {}

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get parameter's value for an instrument's channel.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier.
        """
        self._validate_channel(channel_id=channel_id)

        index = self.dacs.index(channel_id)
        if hasattr(self.settings, parameter.value):
            return getattr(self.settings, parameter.value)[index]
        raise ParameterNotFound(self, parameter)

    @check_device_initialized
    def initial_setup(self):
        """Perform an initial setup."""

        for channel_id in self.dacs:
            self._validate_channel(channel_id=channel_id)

            index = self.dacs.index(channel_id)
            channel = self.device.channel(channel_id)
            channel.dc_mode("fixed")
            channel.output_range(self.span[index])
            channel.output_filter(self.low_pass_filter[index])

            if self.ramping_enabled[index]:
                if (
                    self.ramp_rate[index] < QDevilQDac2._MIN_RAMPING_RATE
                    or self.ramp_rate[index] > QDevilQDac2._MAX_RAMPING_RATE
                ):
                    raise ValueError(
                        f"The ramp rate is out of range on channel {channel_id}. It should be between 0.01 V/s and 2e7 V/s."
                    )
                channel.dc_slew_rate_V_per_s(self.ramp_rate[index])
            else:
                channel.dc_slew_rate_V_per_s(2e7)
            channel.dc_constant_V(0.0)

    @check_device_initialized
    def turn_on(self):
        """Start outputing voltage."""
        for channel_id in self.dacs:
            index = self.dacs.index(channel_id)
            channel = self.device.channel(channel_id)
            channel.dc_constant_V(self.voltage[index])

    @check_device_initialized
    def turn_off(self):
        """Stop outputing voltage."""
        for channel_id in self.dacs:
            channel = self.device.channel(channel_id)
            channel.dc_constant_V(0.0)

    @check_device_initialized
    def reset(self):
        """Reset instrument. This will affect all channels."""
        self.device.reset()

    def _validate_channel(self, channel_id: ChannelID | None):
        """Check if channel identifier is valid and in the allowed range."""
        if channel_id is None:
            raise ValueError(
                "QDevil QDAC-II is a multi-channel instrument. `channel_id` must be specified to get or set parameters."
            )

        channel_id = int(channel_id)
        if channel_id < 1 or channel_id > 24:
            raise ValueError(f"The specified `channel_id`: {channel_id} is out of range. Allowed range is [1, 24].")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the classes used to return the results of the execution of a program.

.. currentmodule:: qililab

Classes
~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~ExperimentResults
    ~result.Result

Functions
~~~~~~~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~stream_results
"""

from .experiment_results import ExperimentResults
from .result import Result
from .stream_results import stream_results

__all__ = ["ExperimentResults", "Result", "stream_results"]



"""Tests for the Keithley2600 class."""
import pytest
from unittest import mock
from qililab.instruments import ParameterNotFound
from qililab.typings import Parameter, InstrumentName
from qililab.instruments.utils import InstrumentFactory
import numpy as np

@pytest.fixture
def keithley2600():
    # Instantiate the Keithley2600 with mocked device and settings
    settings = {
        "alias": "keithley",
        "max_current": 1.0,
        "max_voltage": 10.0
    }
    Keithley2600 = InstrumentFactory.get(InstrumentName.KEITHLEY2600)
    instrument = Keithley2600(settings=settings)
    instrument.device = mock.Mock()
    instrument.device.smua = mock.Mock()
    return instrument


class TestKeithley2600:

    def test_set_parameter_max_current(self, keithley2600):
        # Test setting max current
        keithley2600.set_parameter(Parameter.MAX_CURRENT, 2.0)

        assert keithley2600.max_current == 2.0
        keithley2600.device.smua.limiti.assert_called_with(2.0)

    def test_set_parameter_max_voltage(self, keithley2600):
        # Test setting max voltage
        keithley2600.set_parameter(Parameter.MAX_VOLTAGE, 20.0)

        assert keithley2600.max_voltage == 20.0
        keithley2600.device.smua.limitv.assert_called_with(20.0)

    def test_set_parameter_invalid(self, keithley2600):
        # Test setting an invalid parameter
        with pytest.raises(ParameterNotFound):
            keithley2600.set_parameter("INVALID_PARAMETER", 100)

    def test_initial_setup(self, keithley2600):
        # Test initial setup to ensure it calls the device methods correctly
        keithley2600.initial_setup()

        keithley2600.device.smua.limiti.assert_called_with(keithley2600.max_current)
        keithley2600.device.smua.limitv.assert_called_with(keithley2600.max_voltage)

    def test_turn_on(self, keithley2600):
        # Placeholder test for the turn_on method, which could involve more device actions
        keithley2600.turn_on()
        # Add assertions if turn_on has real behavior in the future

    def test_turn_off(self, keithley2600):
        # Placeholder test for the turn_off method, which could involve more device actions
        keithley2600.turn_off()
        # Add assertions if turn_off has real behavior in the future

    def test_reset(self, keithley2600):
        # Test reset method to ensure it calls device.reset
        keithley2600.reset()
        keithley2600.device.reset.assert_called_once()

    def test_fast_sweep(self, keithley2600):
        # Mock the fast sweep return data
        mock_sweep_data = mock.Mock()
        mock_sweep_data.to_xarray.return_value.to_array.return_value.values.squeeze.return_value = np.array([0.1, 0.2, 0.3])
        keithley2600.device.smua.doFastSweep.return_value = mock_sweep_data

        start, stop, steps = 0.0, 10.0, 3
        x_values, data = keithley2600.fast_sweep(start, stop, steps, mode='IV')

        expected_x_values = np.linspace(start, stop, steps)
        np.testing.assert_array_equal(x_values, expected_x_values)
        np.testing.assert_array_equal(data, np.array([0.1, 0.2, 0.3]))

    def test_max_current_getter(self, keithley2600):
        # Test the max_current getter
        assert keithley2600.max_current == keithley2600.settings.max_current

    def test_max_current_setter(self, keithley2600):
        # Test the max_current setter
        keithley2600.max_current = 5.0
        assert keithley2600.settings.max_current == 5.0
        keithley2600.device.smua.limiti.assert_called_with(5.0)

    def test_max_voltage_getter(self, keithley2600):
        # Test the max_voltage getter
        assert keithley2600.max_voltage == keithley2600.settings.max_voltage

    def test_max_voltage_setter(self, keithley2600):
        # Test the max_voltage setter
        keithley2600.max_voltage = 50.0
        assert keithley2600.settings.max_voltage == 50.0
        keithley2600.device.smua.limitv.assert_called_with(50.0)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime
from pathlib import Path

import h5py
import numpy as np
from ruamel.yaml import YAML

from .platform import Platform
from .settings import Runcard


def save_results(results: np.ndarray, loops: dict[str, np.ndarray], data_path: str, name: str | None = None) -> str:
    """Save the given results and the platform.

    A timestamp is used to create a folder to save the data. The data will be saved within the file located in:
    ``path/yearmonthday/hourminutesecond_name/results.h5``.

    Args:
        results (np.ndarray): Array containing the results to be saved.
        loops (dict): A dictionary containing all the loops used in an experiment. The keys are used to identify the
            loop and the values of the dictionary correspond to the values of the loop.
        data_path (str): Path to the main data directory.
        name (str, optional): Name of the experiment. If given, the name is added to the name of the folder.
            Defaults to None.

    Returns:
        str: Path to folder where the results are saved.

    Examples:
        Imagine you want to run the following sequence:

        .. code-block:: python3

            circuit = Circuit(1)
            circuit.add(gates.X(0))
            circuit.add(gates.M(0))

            results = []
            gain_values = np.arange(0, 1, step=0.01)
            for gain in gain_values:
                platform.set_parameter(alias="drive_q0", parameter=ql.Parameter.GAIN, value=gain)
                result = platform.execute(circuit, num_avg=1000, repetition_duration=200000)
                results.append(result)

            results = np.hstack(results)

        You can then save the results by running:

        .. code-block:: python3

            loops = {"drive_q0_gain": gain_values}
            ql.save_results(results=results, loops=loops, path="data/", name="rabi")

        Imagine we call the cell above on August 22nd of 2023, at 15:14:12. The file will then be saved
        to: ``data/20230822/151412_rabi/results.h5``.
    """
    now = datetime.now()

    # Generate path to the daily folder
    daily_path = Path(data_path) / f"{now.year}{now.month:02d}{now.day:02d}"

    # Check if folder exists, if not create one
    if not os.path.exists(daily_path):
        os.makedirs(daily_path)

    # Generate path to the results folder
    now_path = str(daily_path / f"{now.hour:02d}{now.minute:02d}{now.second:02d}")

    if name is not None:
        now_path = f"{now_path}_{name}"

    # Check if folder exists, if not create one
    if not os.path.exists(now_path):
        os.makedirs(now_path)

    # Create or open an HDF5 file
    with h5py.File(f"{now_path}/results.h5", "w") as hf:
        # Save loops
        g = hf.create_group(name="loops")
        for loop_name, array in loops.items():
            g.create_dataset(name=loop_name, data=array)
        # Save results
        hf.create_dataset("results", data=results)

    return now_path


def load_results(path: str) -> tuple[np.ndarray, dict[str, np.ndarray]]:
    """Load results from the given path to an `.h5` file.

    This function returns a tuple containing the array with the results, and a dictionary containing the loops of
    the experiment.

    Args:
        path (str): Path to the `.h5` file that needs to be loaded.

    Returns:
        tuple[np.ndarray, dict[str, np.ndarray]]: Tuple containing a numpy array with the saved results and a
        dictionary containing the loops of the experiment: ``{"loop_1": loop_1_values, "loop_2": loop_2_values, ...}``.

    Examples:
        Imagine you want to load the results of an experiment located in ``data/20230514/083005/results.h5``:

        >>> results, loops = ql.load_results(path="data/20230514/083005/results.h5")
        >>> results.shape
        (2, 50)
        >>> loops
        {'gain_drive_q0': array([0.  , 0.02, 0.04, 0.06, 0.08, 0.1 , 0.12, 0.14, 0.16, 0.18, 0.2 ,
                0.22, 0.24, 0.26, 0.28, 0.3 , 0.32, 0.34, 0.36, 0.38, 0.4 , 0.42,
                0.44, 0.46, 0.48, 0.5 , 0.52, 0.54, 0.56, 0.58, 0.6 , 0.62, 0.64,
                0.66, 0.68, 0.7 , 0.72, 0.74, 0.76, 0.78, 0.8 , 0.82, 0.84, 0.86,
                0.88, 0.9 , 0.92, 0.94, 0.96, 0.98])}
    """
    with h5py.File(path, "r") as hf:
        loops: dict[str, np.ndarray] = {name: data[:] for name, data in hf["loops"].items()}  # type: ignore
        results = hf["results"][:]  # type: ignore

    return results, loops  # type: ignore


def save_platform(path: str, platform: Platform) -> str:
    """Serialize and save the given platform to the specified path.

    This function saves the cache values of the :class:`.Platform` object during execution as a YAML file.
    It does not read the actual instruments. If you have previously used ``platform.set_parameter()`` without being
    connected to the instruments, it will save this "set" value as the cache values of the :class:`.Platform` object were modified.

    If the `path` string doesn't end with `.yml` or `.yaml`, this function assumes that `path` corresponds to an
    existing folder. The platform will then be saved inside the folder specified by `path` in a file called
    `platform_name.yml`, where `platform_name` corresponds to the `name` attribute of the given `Platform`.


    Args:
        path (str): Path to the folder/file where the YAML file will be saved.
        platform (Platform): Platform class to serialize and save to a YAML file.

    Returns:
        str: Path to the file where the platform is saved.

    Examples:
        If you save a platform by giving the path to a folder:

        >>> ql.save_platform(path="examples/runcards/", platform=platform)

        Qililab will use the name of the platform to create the YAML file. If ``platform.name == "galadriel"``, a file
        will be created in ``examples/runcards/galadriel.yml``.
    """
    if not (path.endswith((".yml", ".yaml"))):
        new_path = Path(path) / f"{platform.name}.yml"
    else:
        new_path = Path(path)

    with open(file=new_path, mode="w", encoding="utf-8") as file:
        YAML().dump(data=platform.to_dict(), stream=file)

    return str(new_path)


def build_platform(runcard: str | dict, new_drivers: bool = False) -> Platform:
    """Builds a :class:`.Platform` object, given a :ref:`runcard <runcards>`.

    Such runcard can be passed in one of the following two ways:
        - a path to a YAML file containing a dictionary of the serialized platform (runcard).
        - directly a dictionary of the serialized platform (runcard).

    |

    The dictionary should follow the next structure:

    .. code-block:: python3

        {
            "name": name,  # str
            "gates_settings": gates_settings,  # dict
            "chip": chip,  # dict
            "buses": buses,  # list[dict]
            "instruments": instruments,  # list[dict]
            "instrument_controllers": instrument_controllers,  # list[dict]
        }

    which contains the information the :class:`.Platform` class uses to connect, setup and control the actual chip, buses and instruments of the laboratory.

    .. note::

        You can find more information about the complete structure of such dictionary, in the :ref:`Runcards <runcards>` section of the documentation.

    Args:
        runcard (str | dict): Path to the platform's runcard YAML file, or direct dictionary of the platform's runcard info.
        new_drivers (bool, optional): Whether to use the new drivers or not. Defaults to False.

    Returns:
        Platform: Platform object.

    Examples:
        Passing the path of a YAML file containing a dictionary of the serialized platform, in the `runcard` argument:

        >>> platform = ql.build_platform(runcard="runcards/galadriel.yml")
        >>> platform.name
        galadriel

        Passing a dictionary of the serialized platform, in the `runcard` argument:

        >>> platform = ql.build_platform(runcard=galadriel_dict)
        >>> platform.name
        galadriel
    """
    if not isinstance(runcard, (str, dict)):
        raise ValueError(
            f"Incorrect type for `runcard` argument in `build_platform()`. Expected (str | dict), got: {type(runcard)}"
        )

    if new_drivers:
        raise NotImplementedError("New drivers are not supported yet.")

    if isinstance(runcard, str):
        with open(file=runcard, mode="r", encoding="utf8") as file:
            yaml = YAML(typ="safe")
            runcard = yaml.load(stream=file)

    runcard_class = Runcard(**runcard)  # type: ignore
    return Platform(runcard=runcard_class)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseDistortion abstract base class."""

from abc import abstractmethod
from copy import deepcopy
from dataclasses import dataclass
from typing import Type, TypeVar

import numpy as np

from qililab.typings.factory_element import FactoryElement
from qililab.utils import Factory

T = TypeVar("T", bound="PulseDistortion")


@dataclass(frozen=True, eq=True, kw_only=True)
class PulseDistortion(FactoryElement):
    """Pulse distortions applied to the :class:`PulseShape`'s envelopes, in order to pre-correct our pulses from future lab physical modifications. ``PulseDistortion`` is their abstract base class.

    Every child of this interface needs to contain an `apply` and `to/from_dict` methods (for serialization).


    The `apply` method will apply the distortion correction to the respective passed envelope, and then will call `normalize_envelope` method of this base class.

    Whenever you call a `PulseDistortion` interface child, apart than their respective arguments you can also pass the `norm_factor` & `auto_norm` arguments, to
    modify how such normalization is done.

    If `self.auto_norm` is True (default) normalizes the resulting envelope to have the same real max height than the starting one. (the max height is the furthest number
    from 0, only checking the real axis/part).

    .. code-block:: python3

        if self.auto_norm:
            auto_norm_envelope = corrected_envelope * (original_norm / corrected_norm)

    If the corrected envelope is zero everywhere or doesn't have a real part (`corrected_norm == 0`) this process is skipped, since the factor in the denominator would diverge:

    .. code-block:: python3

        if self.auto_norm:
            if corrected_norm != 0:
                auto_norm_envelope = corrected_envelope * (original_norm / corrected_norm)

            else:
                auto_norm_envelope = corrected_envelope

    And finally, independently of the `auto_norm`, we will then always apply the manual `self.norm_factor` to the result, reducing the full envelope by its magnitude:

    .. code-block:: python3

        final_envelope = almost_final_envelope * self.norm_factor

    Derived: :class:`BiasTeeCorrection`, :class:`ExponentialCorrection` and :class:`LFilterCorrection`

    Args:
        norm_factor (float): The manual normalization factor that multiplies the envelope in the apply() method. Defaults to 1 (no effect).
        auto_norm (bool): Whether to automatically normalize the corrected envelope with the original max height in the apply() method.
            (the max height is the furthest number from 0 in the envelope, only checking the real axis/part). Default to True.

    Examples:

        Imagine you want to distort a `Rectangular` envelope with an `LFilterCorrection`. You could do:

        >>> from qililab.pulse import Rectangular, LFilterCorrection
        >>> envelope = Rectangular().envelope(duration=50, amplitude=1.0)
        >>> distorted_envelope = LFilterCorrection(a=[0.7, 1.3], b=[0.5, 0.6]).apply(envelope)

        which would return a distorted envelope with the same real max height as the initial.

        >>> np.max(distorted_envelope) == np.max(envelope)
        True

        If instead you wanted to manually modify the envelope to make it 90% smaller than the initial, you would instead do:

        >>> distorted_envelope = LFilterCorrection(a=[0.7, 1.3], b=[0.5, 0.6], norm_factor=0.9).apply(envelope)
        >>> np.max(distorted_envelope) == 0.9 * np.max(envelope)
        True

        And if you wanted to only apply the scipy correction without any normalization, you would then do:

        >>> distorted_envelope_no_norm = LFilterCorrection(a=[0.7, 1.3], b=[0.5, 0.6], auto_norm=False).apply(envelope)
        >>> np.max(distorted_envelope_no_norm) == np.max(envelope)
        False

        Which if ended bigger/smaller than you wanted, you can then also manually modify it like:

        >>> distorted_envelope_manual_norm = LFilterCorrection(
        ...     a=[0.7, 1.3], b=[0.5, 0.6], auto_norm=False, norm_factor=0.8
        ... ).apply(envelope)
        >>> np.max(distorted_envelope_manual_norm) == 0.8 * np.max(distorted_envelope_no_norm) != np.max(envelope)
        True
    """

    norm_factor: float = 1.0  #: Normalization factor.
    auto_norm: bool = True  #: Auto-normalization flag. Defaults to True.

    @abstractmethod
    def apply(self, envelope: np.ndarray) -> np.ndarray:
        """Applies the distortion to the given envelope.

        Args:
            envelope (np.ndarray): Original pulse envelope to be distorted.

        Returns:
            np.ndarray: Distorted pulse envelope.
        """

    @classmethod
    def from_dict(cls: Type[T], dictionary: dict) -> T:
        """Loads a `PulseDistortion` object from a dictionary representation.

        Args:
            dictionary (dict): Dictionary representation of the `PulseDistortion` object, containing all its attributes.
                If called directly from the Abstract parent class: `PulseDistortion`, it must include the child class name to instantiate.

        Returns:
            PulseDistortion: Loaded child class, including the name of the pulse distortion and its attributes set.
        """
        # For calls from parent PulseShape:
        if cls is PulseDistortion:
            shape_class = Factory.get(name=dictionary["name"])
            return shape_class.from_dict(dictionary)

        # For calls from childs directly:
        local_dictionary = deepcopy(dictionary)
        name = local_dictionary.pop("name", None)
        if name not in [cls.name.value, None]:
            raise ValueError(f"Class: {cls.name.value} to instantiate, does not match the given dict name {name}")
        return cls(**local_dictionary)

    def to_dict(self) -> dict:
        """Returns dictionary representation of `PulseDistortion`. This includes the class name an all its attributes.

        Returns:
            dict: Dictionary describing the `PulseDistortion`, including the name and its attributes.
        """
        return {"name": self.name.value} | {
            attribute: getattr(self, attribute) for attribute in self.__dataclass_fields__.keys()
        }

    def normalize_envelope(self, envelope: np.ndarray, corr_envelope: np.ndarray) -> np.ndarray:
        """Normalizes the envelope depending on the `norm_factor` and `auto_norm` attributes.

        If `self.auto_norm` is `True` (default) normalizes the resulting envelope to have the same real max height than the starting one.
        (the max height is the furthest number from 0, only checking the real axis/part)
        If the corrected envelope is zero everywhere or doesn't have a real part this process is skipped.

        Finally it applies the manual `self.norm_factor` to the result, reducing the full envelope by its magnitude.

        For further details on the normalization implementation see the docstring of the class.

        Args:
            envelope (np.ndarray): The original envelope before applying the distortion.
            corr_envelope (np.ndarray): The corrected envelope, but that has not yet been normalized.

        Returns:
            np.ndarray: Normalized final envelope.
        """
        # Automatic normalization if applies
        if self.auto_norm:
            norm = np.max(np.abs(np.real(envelope)))
            corr_norm = np.max(np.abs(np.real(corr_envelope)))

            if corr_norm != 0:
                corr_envelope *= norm / corr_norm

        # Apply manual normalization
        return corr_envelope * self.norm_factor



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from .variable import Domain, Variable


def requires_domain(parameter: str, domain: Domain):
    """Decorator to denote that a parameter requires a variable of a specific domain.

    Args:
        parameter (str): The parameter name.
        domain (Domain): The variable's domain.
    """

    def decorator_function(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Check if the parameter is not inside kwargs, for optional parameters
            if parameter not in kwargs and len(args) == 1:
                kwargs[parameter] = None
            # Get the argument by name
            param_value = kwargs.get(parameter) if parameter in kwargs else None

            if isinstance(param_value, Variable) and param_value.domain != domain:
                raise ValueError(f"Expected domain {domain} for {parameter}, but got {param_value.domain}")
            return func(*args, **kwargs)

        return wrapper

    return decorator_function



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Rohde & Schwarz SGS100A Instrument Controller """
from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.instrument_controller import InstrumentControllerSettings
from qililab.instrument_controllers.single_instrument_controller import SingleInstrumentController
from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instruments.rohde_schwarz.sgs100a import SGS100A
from qililab.typings import RohdeSchwarzSGS100A
from qililab.typings.enums import ConnectionName, InstrumentControllerName, InstrumentTypeName, Parameter


@InstrumentControllerFactory.register
class SGS100AController(SingleInstrumentController):
    """Rohde & Schwarz SGS100A class

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        device (RohdeSchwarz_SGS100A): Instance of the qcodes SGS100A class.
        settings (SGS100ASettings): Settings of the instrument.
    """

    name = InstrumentControllerName.ROHDE_SCHWARZ
    device: RohdeSchwarzSGS100A
    modules: Sequence[SGS100A]

    @dataclass
    class SGS100AControllerSettings(InstrumentControllerSettings):
        """Contains the settings of a specific SGS100A Controller."""

        reference_clock: str

        def __post_init__(self):
            super().__post_init__()
            self.connection.name = ConnectionName.TCP_IP

    settings: SGS100AControllerSettings

    @SingleInstrumentController.CheckConnected
    def initial_setup(self):
        """Initial setup of the instrument."""
        self.device.ref_osc_source(self.settings.reference_clock)
        super().initial_setup()

    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""
        self.device = RohdeSchwarzSGS100A(f"{self.name.value}_{self.alias}", f"TCPIP0::{self.address}::inst0::INSTR")

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, SGS100A):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument is {InstrumentTypeName.ROHDE_SCHWARZ}"
                )

    def to_dict(self):
        """Return a dict representation of the SGS100A controller class."""
        return super().to_dict() | {
            Parameter.REFERENCE_CLOCK.value: self.settings.reference_clock,
        }



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Agilent Vector Network Analyzer E5071B class."""

from dataclasses import dataclass

import numpy as np

from qililab.instruments.decorators import log_set_parameter
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.instruments.vector_network_analyzer import VectorNetworkAnalyzer
from qililab.result.vna_result import VNAResult
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings.instruments.vector_network_analyzer import VectorNetworkAnalyzerDriver


@InstrumentFactory.register
class E5071B(VectorNetworkAnalyzer):
    """Agilent Vector Network Analyzer E5071B"""

    name = InstrumentName.AGILENT_E5071B
    device: VectorNetworkAnalyzerDriver

    @dataclass
    class E5071BSettings(VectorNetworkAnalyzer.VectorNetworkAnalyzerSettings):
        """Contains the settings of a specific VectorNetworkAnalyzer"""

    settings: E5071BSettings

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: int = 1, port: int = 1) -> None:
        """Set instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int): Channel identifier of the parameter to update.
            port (int): Port identifier of the parameter to update.
        """
        channel_id = int(channel_id)
        if parameter == Parameter.POWER:
            value = float(value)
            self.set_power(power=value, channel=channel_id)
            return
        if parameter == Parameter.IF_BANDWIDTH:
            value = float(value)
            self.set_if_bandwidth(value=value, channel=channel_id)
            return
        if parameter == Parameter.ELECTRICAL_DELAY:
            self.electrical_delay = value
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if parameter == Parameter.POWER:
            return self.settings.power
        if parameter == Parameter.IF_BANDWIDTH:
            return self.settings.if_bandwidth
        if parameter == Parameter.ELECTRICAL_DELAY:
            return self.settings.electrical_delay
        raise ParameterNotFound(self, parameter)

    def set_power(self, power: float, channel=1, port=1):
        """Set or read current power"""
        self.settings.power = power
        if self.is_device_active():
            self.send_command(command=f":SOUR{channel}:POW:LEV:IMM:AMPL", arg=f"{power}")

    @VectorNetworkAnalyzer.electrical_delay.setter  # type: ignore
    def electrical_delay(self, time: float):
        """Set electrical delay in channel 1

        Input:
            value (str) : Electrical delay in ns
        """
        self.settings.electrical_delay = time
        if self.is_device_active():
            self.send_command("CALC:MEAS:CORR:EDEL:TIME", f"{time}")

    def set_if_bandwidth(self, value: float, channel=1):
        """Set/query IF Bandwidth for specified channel"""
        self.settings.if_bandwidth = value
        if self.is_device_active():
            self.send_command(command=f":SENS{channel}:BAND:RES", arg=f"{value}")

    def get_data(self):
        """get data"""
        self.send_command(command=":INIT:CONT", arg="OFF")
        self.send_command(command=":INIT:IMM;", arg="*WAI")
        self.send_command(command="CALC:MEAS:DATA:SDATA?", arg="")
        serialized_data = self.read_raw()
        i_0 = serialized_data.find(b"#")
        number_digits = int(serialized_data[i_0 + 1 : i_0 + 2])
        number_bytes = int(serialized_data[i_0 + 2 : i_0 + 2 + number_digits])
        number_data = number_bytes // 4
        number_points = number_data // 2
        v_data = np.frombuffer(
            serialized_data[(i_0 + 2 + number_digits) : (i_0 + 2 + number_digits + number_bytes)],
            dtype=">f",
            count=number_data,
        )
        # data is in I_0,Q0,I1,Q1,I2,Q2,.. format, convert to complex
        measurementsend_commandplex = v_data.reshape((number_points, 2))
        return measurementsend_commandplex[:, 0] + 1j * measurementsend_commandplex[:, 1]

    def acquire_result(self):
        """Convert the data received from the device to a Result object."""
        return VNAResult(data=self.get_data())

    def continuous(self, continuous: bool):
        """set continuous mode
        Args:
            continuous (bool): continuous flag
        """
        arg = "ON" if continuous else "OFF"
        self.send_command(command=":INIT:CONT", arg=arg)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Runcard class."""

from dataclasses import dataclass, field

from qililab.settings.analog.analog_compilation_settings import AnalogCompilationSettings
from qililab.settings.bus_settings import BusSettings
from qililab.settings.digital.digital_compilation_settings import DigitalCompilationSettings


@dataclass
class Runcard:
    """Runcard class. Casts the platform dictionary into a class.

    The input to the constructor should be a dictionary of the desired runcard with the following structure:
    - gates_settings:
    - buses:
    - instruments: List of "instruments" dictionaries
    - instrument_controllers: List of "instrument_controllers" dictionaries

    The gates_settings and bus dictionaries will be passed to their corresponding Runcard.GatesSettings and Runcard.Bus
    classes here, meanwhile the instruments and instrument_controllers will remain dictionaries.

    Then this full class gets passed to the Platform who will instantiate the actual Buses/Bus and the
    corresponding Instrument classes with the settings attributes of this class.

    Args:
        gates_settings (dict): Gates settings dictionary -> Runcard.GatesSettings inner dataclass
        buses (list[dict]): List of Bus settings dictionaries -> list[Runcard.Bus] settings inner dataclass
        instruments (list[dict]): List of dictionaries containing the "instruments" information (does not transform)
        instruments_controllers (list[dict]): List of dictionaries containing the "instrument_controllers" information
            (does not transform)
    """

    name: str
    instruments: list[dict] = field(default_factory=list)
    instrument_controllers: list[dict] = field(default_factory=list)
    buses: list[BusSettings] = field(default_factory=list)
    digital: DigitalCompilationSettings | None = None
    analog: AnalogCompilationSettings | None = None

    def __post_init__(self):
        self.buses = [BusSettings(**bus) for bus in self.buses]
        self.digital = DigitalCompilationSettings(**self.digital) if self.digital is not None else None
        self.analog = AnalogCompilationSettings(**self.analog) if self.analog is not None else None



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox QCM class"""

from dataclasses import dataclass

from qililab.instruments.qblox.qblox_module import QbloxModule
from qililab.instruments.utils import InstrumentFactory
from qililab.typings.enums import InstrumentName


@InstrumentFactory.register
class QbloxQCM(QbloxModule):
    """Qblox QCM class.

    Args:
        settings (QBloxQCMSettings): Settings of the instrument.
    """

    name = InstrumentName.QBLOX_QCM

    @dataclass
    class QbloxQCMSettings(QbloxModule.QbloxModuleSettings):
        """Contains the settings of a specific pulsar."""

    settings: QbloxQCMSettings

    def is_awg(self) -> bool:
        """Returns True if instrument is an AWG."""
        return True
